#
# Base solver class
#
import copy
import itertools
from scipy.sparse import block_diag
import multiprocessing as mp
import numbers
import sys
import warnings

import casadi
import numpy as np

import pybamm
from pybamm.expression_tree.binary_operators import _Heaviside


class BaseSolver(object):
    """Solve a discretised model.

    Parameters
    ----------
    method : str, optional
        The method to use for integration, specific to each solver
    rtol : float, optional
        The relative tolerance for the solver (default is 1e-6).
    atol : float, optional
        The absolute tolerance for the solver (default is 1e-6).
    root_method : str or pybamm algebraic solver class, optional
        The method to use to find initial conditions (for DAE solvers).
        If a solver class, must be an algebraic solver class.
        If "casadi",
        the solver uses casadi's Newton rootfinding algorithm to find initial
        conditions. Otherwise, the solver uses 'scipy.optimize.root' with method
        specified by 'root_method' (e.g. "lm", "hybr", ...)
    root_tol : float, optional
        The tolerance for the initial-condition solver (default is 1e-6).
    extrap_tol : float, optional
        The tolerance to assert whether extrapolation occurs or not. Default is 0.
    """

    def __init__(
        self,
        method=None,
        rtol=1e-6,
        atol=1e-6,
        root_method=None,
        root_tol=1e-6,
        extrap_tol=0,
        max_steps="deprecated",
    ):
        self.method = method
        self.rtol = rtol
        self.atol = atol
        self.root_tol = root_tol
        self.root_method = root_method
        self.extrap_tol = extrap_tol
        if max_steps != "deprecated":
            raise ValueError(
                "max_steps has been deprecated, and should be set using the "
                "solver-specific extra-options dictionaries instead"
            )
        self.models_set_up = {}

        # Defaults, can be overwritten by specific solver
        self.name = "Base solver"
        self.ode_solver = False
        self.algebraic_solver = False

    @property
    def root_method(self):
        return self._root_method

    @root_method.setter
    def root_method(self, method):
        if method == "casadi":
            method = pybamm.CasadiAlgebraicSolver(self.root_tol)
        elif isinstance(method, str):
            method = pybamm.AlgebraicSolver(method, self.root_tol)
        elif not (
            method is None
            or (
                isinstance(method, pybamm.BaseSolver)
                and method.algebraic_solver is True
            )
        ):
            raise pybamm.SolverError("Root method must be an algebraic solver")
        self._root_method = method

    def copy(self):
        """Returns a copy of the solver"""
        new_solver = copy.copy(self)
        # clear models_set_up
        new_solver.models_set_up = {}
        return new_solver

    def set_up(self, model, inputs=None, t_eval=None, ics_only=False):
        """Unpack model, perform checks, and calculate jacobian.

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The model whose solution to calculate. Must have attributes rhs and
            initial_conditions
        inputs : dict, optional
            Any input parameters to pass to the model when solving
        t_eval : numeric type, optional
            The times (in seconds) at which to compute the solution
        """
        pybamm.logger.info("Start solver set-up")

        # Check model.algebraic for ode solvers
        if self.ode_solver is True and len(model.algebraic) > 0:
            raise pybamm.SolverError(
                "Cannot use ODE solver '{}' to solve DAE model".format(self.name)
            )
        # Check model.rhs for algebraic solvers
        if self.algebraic_solver is True and len(model.rhs) > 0:
            raise pybamm.SolverError(
                """Cannot use algebraic solver to solve model with time derivatives"""
            )
        # casadi solver won't allow solving algebraic model so we have to raise an
        # error here
        if isinstance(self, pybamm.CasadiSolver) and len(model.rhs) == 0:
            raise pybamm.SolverError(
                "Cannot use CasadiSolver to solve algebraic model, "
                "use CasadiAlgebraicSolver instead"
            )
        # Discretise model if it isn't already discretised
        # This only works with purely 0D models, as otherwise the mesh and spatial
        # method should be specified by the user
        if model.is_discretised is False:
            try:
                disc = pybamm.Discretisation()
                disc.process_model(model)
            except pybamm.DiscretisationError as e:
                raise pybamm.DiscretisationError(
                    "Cannot automatically discretise model, "
                    "model should be discretised before solving ({})".format(e)
                )

        inputs = inputs or {}

        # Set model timescale
        model.timescale_eval = model.timescale.evaluate(inputs=inputs)
        # Set model lengthscales
        model.length_scales_eval = {
            domain: scale.evaluate(inputs=inputs)
            for domain, scale in model.length_scales.items()
        }
        if (
            isinstance(self, (pybamm.CasadiSolver, pybamm.CasadiAlgebraicSolver))
        ) and model.convert_to_format != "casadi":
            pybamm.logger.warning(
                "Converting {} to CasADi for solving with CasADi solver".format(
                    model.name
                )
            )
            model.convert_to_format = "casadi"
        if (
            isinstance(self.root_method, pybamm.CasadiAlgebraicSolver)
            and model.convert_to_format != "casadi"
        ):
            pybamm.logger.warning(
                "Converting {} to CasADi for calculating ICs with CasADi".format(
                    model.name
                )
            )
            model.convert_to_format = "casadi"

        # find all the input parameters in the model
        input_parameters = {}
        for equation in [
            model.concatenated_rhs,
            model.concatenated_algebraic,
            model.concatenated_initial_conditions,
        ]:
            input_parameters.update(
                {
                    symbol._id: symbol
                    for symbol in equation.pre_order()
                    if isinstance(symbol, pybamm.InputParameter)
                }
            )

        # set default calculate sensitivities on model
        if not hasattr(model, "calculate_sensitivities"):
            model.calculate_sensitivities = []

        # see if we need to form the explicit sensitivity equations
        calculate_sensitivities_explicit = False
        if model.calculate_sensitivities and not isinstance(self, pybamm.IDAKLUSolver):
            calculate_sensitivities_explicit = True

        # if we are calculating sensitivities explicitly then the number of
        # equations will change
        if calculate_sensitivities_explicit:
            num_parameters = 0
            for name in model.calculate_sensitivities:
                # if not a number, assume its a vector
                if isinstance(inputs[name], numbers.Number):
                    num_parameters += 1
                else:
                    num_parameters += len(inputs[name])
            model.len_rhs_sens = model.len_rhs * num_parameters
            model.len_alg_sens = model.len_alg * num_parameters

        if model.convert_to_format != "casadi":
            # Create Jacobian from concatenated rhs and algebraic
            y = pybamm.StateVector(slice(0, model.len_rhs_and_alg))
            # set up Jacobian object, for re-use of dict
            jacobian = pybamm.Jacobian()

        else:
            # Convert model attributes to casadi
            t_casadi = casadi.MX.sym("t")
            # Create the symbolic state vectors
            y_diff = casadi.MX.sym("y_diff", model.len_rhs)
            y_alg = casadi.MX.sym("y_alg", model.len_alg)
            y_casadi = casadi.vertcat(y_diff, y_alg)
            p_casadi = {}
            for name, value in inputs.items():
                if isinstance(value, numbers.Number):
                    p_casadi[name] = casadi.MX.sym(name)
                else:
                    p_casadi[name] = casadi.MX.sym(name, value.shape[0])
            p_casadi_stacked = casadi.vertcat(*[p for p in p_casadi.values()])
            # sensitivity vectors
            if calculate_sensitivities_explicit:
                pS_casadi_stacked = casadi.vertcat(
                    *[p_casadi[name] for name in model.calculate_sensitivities]
                )
                S_x = casadi.MX.sym("S_x", model.len_rhs_sens)
                S_z = casadi.MX.sym("S_z", model.len_alg_sens)
                y_and_S = casadi.vertcat(y_diff, S_x, y_alg, S_z)
            else:
                y_and_S = y_casadi

        def process(func, name, use_jacobian=None):
            def report(string):
                # don't log event conversion
                if "event" not in string:
                    pybamm.logger.verbose(string)

            if use_jacobian is None:
                use_jacobian = model.use_jacobian

            if model.convert_to_format == "jax":
                report(f"Converting {name} to jax")
                func = pybamm.EvaluatorJax(func)
                jacp = None
                if model.calculate_sensitivities:
                    report(
                        (
                            f"Calculating sensitivities for {name} with respect "
                            f"to parameters {model.calculate_sensitivities} using jax"
                        )
                    )
                    jacp = func.get_sensitivities()
                    jacp = jacp.evaluate
                if use_jacobian:
                    report(f"Calculating jacobian for {name} using jax")
                    jac = func.get_jacobian()
                    jac = jac.evaluate
                else:
                    jac = None

                func = func.evaluate

            elif model.convert_to_format != "casadi":
                # Process with pybamm functions, optionally converting
                # to python evaluator
                if model.calculate_sensitivities:
                    report(
                        (
                            f"Calculating sensitivities for {name} with respect "
                            f"to parameters {model.calculate_sensitivities}"
                        )
                    )
                    jacp_dict = {
                        p: func.diff(pybamm.InputParameter(p))
                        for p in model.calculate_sensitivities
                    }
                    if model.convert_to_format == "python":
                        report(f"Converting sensitivities for {name} to python")
                        jacp_dict = {
                            p: pybamm.EvaluatorPython(jacp)
                            for p, jacp in jacp_dict.items()
                        }

                    # jacp should be a function that returns a dict of sensitivities
                    def jacp(*args, **kwargs):
                        return {
                            k: v.evaluate(*args, **kwargs) for k, v in jacp_dict.items()
                        }

                else:
                    jacp = None

                if use_jacobian:
                    report(f"Calculating jacobian for {name}")
                    jac = jacobian.jac(func, y)
                    if model.convert_to_format == "python":
                        report(f"Converting jacobian for {name} to python")
                        jac = pybamm.EvaluatorPython(jac)
                    jac = jac.evaluate
                else:
                    jac = None

                if model.convert_to_format == "python":
                    report(f"Converting {name} to python")
                    func = pybamm.EvaluatorPython(func)

                func = func.evaluate

            else:
                # Process with CasADi
                report(f"Converting {name} to CasADi")
                func = func.to_casadi(t_casadi, y_casadi, inputs=p_casadi)
                # Add sensitivity vectors to the rhs and algebraic equations
                jacp = None
                if calculate_sensitivities_explicit:
                    # The formulation is as per Park, S., Kato, D., Gima, Z., Klein, R.,
                    # & Moura, S. (2018).  Optimal experimental design for
                    # parameterization of an electrochemical lithium-ion battery model.
                    # Journal of The Electrochemical Society, 165(7), A1309.". See #1100
                    # for details
                    if name == "RHS" and model.len_rhs > 0:
                        report(
                            "Creating explicit forward sensitivity equations "
                            "for rhs using CasADi"
                        )
                        df_dx = casadi.jacobian(func, y_diff)
                        df_dp = casadi.jacobian(func, pS_casadi_stacked)
                        S_x_mat = S_x.reshape(
                            (model.len_rhs, pS_casadi_stacked.shape[0])
                        )
                        if model.len_alg == 0:
                            S_rhs = (df_dx @ S_x_mat + df_dp).reshape((-1, 1))
                        else:
                            df_dz = casadi.jacobian(func, y_alg)
                            S_z_mat = S_z.reshape(
                                (model.len_alg, pS_casadi_stacked.shape[0])
                            )
                            S_rhs = (df_dx @ S_x_mat + df_dz @ S_z_mat + df_dp).reshape(
                                (-1, 1)
                            )
                        func = casadi.vertcat(func, S_rhs)
                    if name == "algebraic" and model.len_alg > 0:
                        report(
                            "Creating explicit forward sensitivity equations "
                            "for algebraic using CasADi"
                        )
                        dg_dz = casadi.jacobian(func, y_alg)
                        dg_dp = casadi.jacobian(func, pS_casadi_stacked)
                        S_z_mat = S_z.reshape(
                            (model.len_alg, pS_casadi_stacked.shape[0])
                        )
                        if model.len_rhs == 0:
                            S_alg = (dg_dz @ S_z_mat + dg_dp).reshape((-1, 1))
                        else:
                            dg_dx = casadi.jacobian(func, y_diff)
                            S_x_mat = S_x.reshape(
                                (model.len_rhs, pS_casadi_stacked.shape[0])
                            )
                            S_alg = (dg_dx @ S_x_mat + dg_dz @ S_z_mat + dg_dp).reshape(
                                (-1, 1)
                            )
                        func = casadi.vertcat(func, S_alg)
                    if name == "initial_conditions":
                        if model.len_rhs == 0 or model.len_alg == 0:
                            S_0 = casadi.jacobian(func, pS_casadi_stacked).reshape(
                                (-1, 1)
                            )
                            func = casadi.vertcat(func, S_0)
                        else:
                            x0 = func[: model.len_rhs]
                            z0 = func[model.len_rhs :]
                            Sx_0 = casadi.jacobian(x0, pS_casadi_stacked).reshape(
                                (-1, 1)
                            )
                            Sz_0 = casadi.jacobian(z0, pS_casadi_stacked).reshape(
                                (-1, 1)
                            )
                            func = casadi.vertcat(x0, Sx_0, z0, Sz_0)
                elif model.calculate_sensitivities:
                    report(
                        (
                            f"Calculating sensitivities for {name} with respect "
                            f"to parameters {model.calculate_sensitivities} using "
                            "CasADi"
                        )
                    )
                    jacp_dict = {}
                    for pname in model.calculate_sensitivities:
                        p_diff = casadi.jacobian(func, p_casadi[pname])
                        jacp_dict[pname] = casadi.Function(
                            name, [t_casadi, y_casadi, p_casadi_stacked], [p_diff]
                        )
                    # jacp should be a function that returns a dict of sensitivities

                    def jacp(*args, **kwargs):
                        return {k: v(*args, **kwargs) for k, v in jacp_dict.items()}

                if use_jacobian:
                    report(f"Calculating jacobian for {name} using CasADi")
                    jac_casadi = casadi.jacobian(func, y_and_S)
                    jac = casadi.Function(
                        name, [t_casadi, y_and_S, p_casadi_stacked], [jac_casadi]
                    )
                else:
                    jac = None

                func = casadi.Function(
                    name, [t_casadi, y_and_S, p_casadi_stacked], [func]
                )
            if name == "residuals":
                func_call = Residuals(func, name, model)
            else:
                func_call = SolverCallable(func, name, model)
            if jac is not None:
                jac_call = SolverCallable(jac, name + "_jac", model)
            else:
                jac_call = None
            if jacp is not None:
                jacp_call = SensitivityCallable(
                    jacp,
                    name + "_sensitivity_wrt_inputs",
                    model,
                    model.convert_to_format,
                )
            else:
                jacp_call = None
            return func, func_call, jac_call, jacp_call

        # Process initial conditions
        initial_conditions = process(
            model.concatenated_initial_conditions,
            "initial_conditions",
            use_jacobian=False,
        )[0]
        init_eval = InitialConditions(initial_conditions, model)

        if calculate_sensitivities_explicit:
            y0_total_size = (
                model.len_rhs + model.len_rhs_sens + model.len_alg + model.len_alg_sens
            )
            init_eval.y_dummy = np.zeros((y0_total_size, 1))
        else:
            init_eval.y_dummy = np.zeros((model.len_rhs_and_alg, 1))

        # Calculate initial conditions
        model.y0 = init_eval(inputs)
        model.init_eval = init_eval

        if not ics_only:
            # Check for heaviside and modulo functions in rhs and algebraic and add
            # discontinuity events if these exist.
            # Note: only checks for the case of t < X, t <= X, X < t, or X <= t,
            # but also accounts for the fact that t might be dimensional
            # Only do this for DAE models as ODE models can deal with discontinuities
            # fine
            if len(model.algebraic) > 0:
                for symbol in itertools.chain(
                    model.concatenated_rhs.pre_order(),
                    model.concatenated_algebraic.pre_order(),
                ):
                    if isinstance(symbol, _Heaviside):
                        found_t = False
                        # Dimensionless
                        if symbol.right.id == pybamm.t.id:
                            expr = symbol.left
                            found_t = True
                        elif symbol.left.id == pybamm.t.id:
                            expr = symbol.right
                            found_t = True
                        # Dimensional
                        elif symbol.right.id == (pybamm.t * model.timescale_eval).id:
                            expr = (
                                symbol.left.new_copy() / symbol.right.right.new_copy()
                            )
                            found_t = True
                        elif symbol.left.id == (pybamm.t * model.timescale_eval).id:
                            expr = (
                                symbol.right.new_copy() / symbol.left.right.new_copy()
                            )
                            found_t = True

                        # Update the events if the heaviside function depended on t
                        if found_t:
                            model.events.append(
                                pybamm.Event(
                                    str(symbol),
                                    expr.new_copy(),
                                    pybamm.EventType.DISCONTINUITY,
                                )
                            )
                    elif isinstance(symbol, pybamm.Modulo):
                        found_t = False
                        # Dimensionless
                        if symbol.left.id == pybamm.t.id:
                            expr = symbol.right
                            found_t = True
                        # Dimensional
                        elif symbol.left.id == (pybamm.t * model.timescale_eval).id:
                            expr = (
                                symbol.right.new_copy() / symbol.left.right.new_copy()
                            )
                            found_t = True

                        # Update the events if the modulo function depended on t
                        if found_t:
                            if t_eval is None:
                                N_events = 200
                            else:
                                N_events = t_eval[-1] // expr.value

                            for i in np.arange(N_events):
                                model.events.append(
                                    pybamm.Event(
                                        str(symbol),
                                        expr.new_copy() * pybamm.Scalar(i + 1),
                                        pybamm.EventType.DISCONTINUITY,
                                    )
                                )

            # Process rhs, algebraic and event expressions
            rhs, rhs_eval, jac_rhs, jacp_rhs = process(model.concatenated_rhs, "RHS")
            algebraic, algebraic_eval, jac_algebraic, jacp_algebraic = process(
                model.concatenated_algebraic, "algebraic"
            )
            casadi_terminate_events = []
            terminate_events_eval = []
            interpolant_extrapolation_events_eval = []
            discontinuity_events_eval = []
            for n, event in enumerate(model.events):
                if event.event_type == pybamm.EventType.DISCONTINUITY:
                    # discontinuity events are evaluated before the solver is called,
                    # so don't need to process them
                    discontinuity_events_eval.append(event)
                elif event.event_type == pybamm.EventType.SWITCH:
                    if (
                        isinstance(self, pybamm.CasadiSolver)
                        and self.mode == "fast with events"
                        and model.algebraic != {}
                    ):
                        # Save some events to casadi_terminate_events for the 'fast with
                        # events' mode of the casadi solver
                        # We only need to do this if the model is a DAE model
                        # see #1082
                        k = 20
                        init_sign = float(
                            np.sign(event.evaluate(0, model.y0.full(), inputs=inputs))
                        )
                        # We create a sigmoid for each event which will multiply the
                        # rhs. Doing * 2 - 1 ensures that when the event is crossed,
                        # the sigmoid is zero. Hence the rhs is zero and the solution
                        # stays constant for the rest of the simulation period
                        # We can then cut off the part after the event was crossed
                        event_sigmoid = (
                            pybamm.sigmoid(0, init_sign * event.expression, k) * 2 - 1
                        )
                        event_casadi = process(
                            event_sigmoid, f"event_{n}", use_jacobian=False
                        )[0]
                        # use the actual casadi object as this will go into the rhs
                        casadi_terminate_events.append(event_casadi)
                else:
                    # use the function call
                    event_eval = process(
                        event.expression, f"event_{n}", use_jacobian=False
                    )[1]
                    if event.event_type == pybamm.EventType.TERMINATION:
                        terminate_events_eval.append(event_eval)
                    elif event.event_type == pybamm.EventType.INTERPOLANT_EXTRAPOLATION:
                        interpolant_extrapolation_events_eval.append(event_eval)

            # Add the solver attributes
            model.rhs_eval = rhs_eval
            model.algebraic_eval = algebraic_eval
            model.jac_algebraic_eval = jac_algebraic
            model.casadi_terminate_events = casadi_terminate_events
            model.terminate_events_eval = terminate_events_eval
            model.discontinuity_events_eval = discontinuity_events_eval
            model.interpolant_extrapolation_events_eval = (
                interpolant_extrapolation_events_eval
            )

            # if we have changed the equations to include the explicit sensitivity
            # equations, then we also need to update the mass matrix and bounds
            if calculate_sensitivities_explicit:
                if model.len_rhs != 0:
                    n_inputs = model.len_rhs_sens // model.len_rhs
                elif model.len_alg != 0:
                    n_inputs = model.len_alg_sens // model.len_alg
                if model.bounds[0].shape[0] == model.len_rhs_and_alg:
                    model.bounds = (
                        np.repeat(model.bounds[0], n_inputs + 1),
                        np.repeat(model.bounds[1], n_inputs + 1),
                    )
                if (
                    model.mass_matrix is not None
                    and model.mass_matrix.shape[0] == model.len_rhs_and_alg
                ):

                    if model.mass_matrix_inv is not None:
                        model.mass_matrix_inv = pybamm.Matrix(
                            block_diag(
                                [model.mass_matrix_inv.entries] * (n_inputs + 1),
                                format="csr",
                            )
                        )
                    model.mass_matrix = pybamm.Matrix(
                        block_diag(
                            [model.mass_matrix.entries] * (n_inputs + 1), format="csr"
                        )
                    )
            else:
                # take care if calculate_sensitivites used then not used
                if model.bounds[0].shape[0] > model.len_rhs_and_alg:
                    model.bounds = (
                        model.bounds[0][: model.len_rhs_and_alg],
                        model.bounds[1][: model.len_rhs_and_alg],
                    )
                if (
                    model.mass_matrix is not None
                    and model.mass_matrix.shape[0] > model.len_rhs_and_alg
                ):
                    if model.mass_matrix_inv is not None:
                        model.mass_matrix_inv = pybamm.Matrix(
                            model.mass_matrix_inv.entries[
                                : model.len_rhs, : model.len_rhs
                            ]
                        )
                    model.mass_matrix = pybamm.Matrix(
                        model.mass_matrix.entries[
                            : model.len_rhs_and_alg, : model.len_rhs_and_alg
                        ]
                    )

            # Save CasADi functions for the CasADi solver
            # Note: when we pass to casadi the ode part of the problem must be in
            # explicit form so we pre-multiply by the inverse of the mass matrix
            if isinstance(self.root_method, pybamm.CasadiAlgebraicSolver) or isinstance(
                self, (pybamm.CasadiSolver, pybamm.CasadiAlgebraicSolver)
            ):
                # can use DAE solver to solve model with algebraic equations only
                if len(model.rhs) > 0:
                    mass_matrix_inv = casadi.MX(model.mass_matrix_inv.entries)
                    explicit_rhs = mass_matrix_inv @ rhs(
                        t_casadi, y_and_S, p_casadi_stacked
                    )
                    model.casadi_rhs = casadi.Function(
                        "rhs", [t_casadi, y_and_S, p_casadi_stacked], [explicit_rhs]
                    )
                model.casadi_algebraic = algebraic
                model.casadi_sensitivities_rhs = jacp_rhs
                model.casadi_sensitivities_algebraic = jacp_algebraic
            if len(model.rhs) == 0:
                # No rhs equations: residuals is algebraic only
                model.residuals_eval = Residuals(algebraic, "residuals", model)
                model.jacobian_eval = jac_algebraic
                model.sensitivities_eval = jacp_algebraic
            elif len(model.algebraic) == 0:
                # No algebraic equations: residuals is rhs only
                model.residuals_eval = Residuals(rhs, "residuals", model)
                model.jacobian_eval = jac_rhs
                model.sensitivities_eval = jacp_rhs
            # Calculate consistent initial conditions for the algebraic equations
            else:
                all_states = pybamm.NumpyConcatenation(
                    model.concatenated_rhs, model.concatenated_algebraic
                )
                # Process again, uses caching so should be quick
                residuals_eval, jacobian_eval, jacobian_wrtp_eval = process(
                    all_states, "residuals"
                )[1:]
                model.residuals_eval = residuals_eval
                model.jacobian_eval = jacobian_eval
                model.sensitivities_eval = jacobian_wrtp_eval

        pybamm.logger.info("Finish solver set-up")

    def _set_initial_conditions(self, model, inputs, update_rhs):
        """
        Set initial conditions for the model. This is skipped if the solver is an
        algebraic solver (since this would make the algebraic solver redundant), and if
        the model doesn't have any algebraic equations (since there are no initial
        conditions to be calculated in this case).

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The model for which to calculate initial conditions.
        inputs : dict
            Any input parameters to pass to the model when solving
        update_rhs : bool
            Whether to update the rhs. True for 'solve', False for 'step'.

        """

        if self.algebraic_solver is True:
            # Don't update model.y0
            return None
        elif len(model.algebraic) == 0:
            if update_rhs is True:
                # Recalculate initial conditions for the rhs equations
                y0 = model.init_eval(inputs)
            else:
                # Don't update model.y0
                return None
        else:
            if update_rhs is True:
                # Recalculate initial conditions for the rhs equations
                y0_from_inputs = model.init_eval(inputs)
                # Reuse old solution for algebraic equations
                y0_from_model = model.y0
                len_rhs = model.len_rhs
                # update model.y0, which is used for initialising the algebraic solver
                if len_rhs == 0:
                    model.y0 = y0_from_model
                else:
                    model.y0 = casadi.vertcat(
                        y0_from_inputs[:len_rhs], y0_from_model[len_rhs:]
                    )
            y0 = self.calculate_consistent_state(model, 0, inputs)
        # Make y0 a function of inputs if doing symbolic with casadi
        model.y0 = y0

    def calculate_consistent_state(self, model, time=0, inputs=None):
        """
        Calculate consistent state for the algebraic equations through
        root-finding. model.y0 is used as the initial guess for rootfinding

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The model for which to calculate initial conditions.
        time : float
            The time at which to calculate the states
        inputs_dict : dict, optional
            Any input parameters to pass to the model when solving

        Returns
        -------
        y0_consistent : array-like, same shape as y0_guess
            Initial conditions that are consistent with the algebraic equations (roots
            of the algebraic equations). If self.root_method == None then returns
            model.y0.
        """
        pybamm.logger.debug("Start calculating consistent states")
        if self.root_method is None:
            return model.y0
        try:
            root_sol = self.root_method._integrate(model, np.array([time]), inputs)
        except pybamm.SolverError as e:
            raise pybamm.SolverError(
                "Could not find consistent states: {}".format(e.args[0])
            )
        pybamm.logger.debug("Found consistent states")

        y0 = root_sol.all_ys[0]
        if isinstance(y0, np.ndarray):
            y0 = y0.flatten()
        return y0

    def solve(
        self,
        model,
        t_eval=None,
        external_variables=None,
        inputs=None,
        initial_conditions=None,
        nproc=None,
        calculate_sensitivities=False,
    ):
        """
        Execute the solver setup and calculate the solution of the model at
        specified times.

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The model whose solution to calculate. Must have attributes rhs and
            initial_conditions
        t_eval : numeric type
            The times (in seconds) at which to compute the solution
        external_variables : dict
            A dictionary of external variables and their corresponding
            values at the current time
        inputs : dict or list, optional
            A dictionary or list of dictionaries describing any input parameters to
            pass to the model when solving
        initial_conditions : :class:`pybamm.Symbol`, optional
            Initial conditions to use when solving the model. If None (default),
            `model.concatenated_initial_conditions` is used. Otherwise, must be a symbol
            of size `len(model.rhs) + len(model.algebraic)`.
        nproc : int, optional
            Number of processes to use when solving for more than one set of input
            parameters. Defaults to value returned by "os.cpu_count()".
        calculate_sensitivites : list of str or bool
            If true, solver calculates sensitivities of all input parameters.
            If only a subset of sensitivities are required, can also pass a
            list of input parameter names

        Returns
        -------
        :class:`pybamm.Solution` or list of :class:`pybamm.Solution` objects.
             If type of `inputs` is `list`, return a list of corresponding
             :class:`pybamm.Solution` objects.

        Raises
        ------
        :class:`pybamm.ModelError`
            If an empty model is passed (`model.rhs = {}` and `model.algebraic={}` and
            `model.variables = {}`)

        """
        pybamm.logger.info("Start solving {} with {}".format(model.name, self.name))

        # get a list-only version of calculate_sensitivities
        if isinstance(calculate_sensitivities, bool):
            if calculate_sensitivities:
                calculate_sensitivities_list = [p for p in inputs.keys()]
            else:
                calculate_sensitivities_list = []
        else:
            calculate_sensitivities_list = calculate_sensitivities

        # Make sure model isn't empty
        if len(model.rhs) == 0 and len(model.algebraic) == 0:
            if not isinstance(self, pybamm.DummySolver):
                raise pybamm.ModelError(
                    "Cannot solve empty model, use `pybamm.DummySolver` instead"
                )

        # t_eval can only be None if the solver is an algebraic solver. In that case
        # set it to 0
        if t_eval is None:
            if self.algebraic_solver is True:
                t_eval = np.array([0])
            else:
                raise ValueError("t_eval cannot be None")

        # If t_eval is provided as [t0, tf] return the solution at 100 points
        elif isinstance(t_eval, list):
            if len(t_eval) == 1 and self.algebraic_solver is True:
                pass
            elif len(t_eval) != 2:
                raise pybamm.SolverError(
                    "'t_eval' can be provided as an array of times at which to "
                    "return the solution, or as a list [t0, tf] where t0 is the "
                    "initial time and tf is the final time, but has been provided "
                    "as a list of length {}.".format(len(t_eval))
                )
            else:
                t_eval = np.linspace(t_eval[0], t_eval[-1], 100)

        # Make sure t_eval is monotonic
        if (np.diff(t_eval) < 0).any():
            raise pybamm.SolverError("t_eval must increase monotonically")

        # Set up external variables and inputs
        #
        # Argument "inputs" can be either a list of input dicts or
        # a single dict. The remaining of this function is only working
        # with variable "input_list", which is a list of dictionaries.
        # If "inputs" is a single dict, "inputs_list" is a list of only one dict.
        inputs_list = inputs if isinstance(inputs, list) else [inputs]
        ext_and_inputs_list = [
            self._set_up_ext_and_inputs(
                model, external_variables, inputs, calculate_sensitivities_list
            )
            for inputs in inputs_list
        ]

        # Cannot use multiprocessing with model in "jax" format
        if (len(inputs_list) > 1) and model.convert_to_format == "jax":
            raise pybamm.SolverError(
                "Cannot solve list of inputs with multiprocessing "
                'when model in format "jax".'
            )

        # Check that calculate_sensitivites have not been updated
        calculate_sensitivities_list.sort()
        if not hasattr(model, "calculate_sensitivities"):
            model.calculate_sensitivities = []
        model.calculate_sensitivities.sort()
        if calculate_sensitivities_list != model.calculate_sensitivities:
            self.models_set_up.pop(model, None)
            # CasadiSolver caches its integrators using model, so delete this too
            if isinstance(self, pybamm.CasadiSolver):
                self.integrators.pop(model, None)

        # save sensitivity parameters so we can identify them later on
        # (FYI: this is used in the Solution class)
        model.calculate_sensitivities = calculate_sensitivities_list

        # Set up (if not done already)
        timer = pybamm.Timer()
        if model not in self.models_set_up:
            # It is assumed that when len(inputs_list) > 1, model set
            # up (initial condition, time-scale and length-scale) does
            # not depend on input parameters. Thefore only `ext_and_inputs[0]`
            # is passed to `set_up`.
            # See https://github.com/pybamm-team/PyBaMM/pull/1261
            self.set_up(model, ext_and_inputs_list[0], t_eval)
            self.models_set_up.update(
                {model: {"initial conditions": model.concatenated_initial_conditions}}
            )
        else:
            ics_set_up = self.models_set_up[model]["initial conditions"]
            # Check that initial conditions have not been updated
            if ics_set_up.id != model.concatenated_initial_conditions.id:
                # If the new initial conditions are different, set up again
                self.set_up(model, ext_and_inputs_list[0], t_eval, ics_only=True)
                self.models_set_up[model][
                    "initial conditions"
                ] = model.concatenated_initial_conditions

        set_up_time = timer.time()
        timer.reset()

        # (Re-)calculate consistent initial conditions
        # Assuming initial conditions do not depend on input parameters
        # when len(inputs_list) > 1, only `ext_and_inputs_list[0]`
        # is passed to `_set_initial_conditions`.
        # See https://github.com/pybamm-team/PyBaMM/pull/1261
        if len(inputs_list) > 1:
            all_inputs_names = set(
                itertools.chain.from_iterable(
                    [ext_and_inputs.keys() for ext_and_inputs in ext_and_inputs_list]
                )
            )
            initial_conditions_node_names = set(
                [it.name for it in model.concatenated_initial_conditions.pre_order()]
            )
            if all_inputs_names.issubset(initial_conditions_node_names):
                raise pybamm.SolverError(
                    "Input parameters cannot appear in expression "
                    "for initial conditions."
                )

        self._set_initial_conditions(model, ext_and_inputs_list[0], update_rhs=True)

        # Non-dimensionalise time
        t_eval_dimensionless = t_eval / model.timescale_eval

        # Calculate discontinuities
        discontinuities = [
            # Assuming that discontinuities do not depend on
            # input parameters when len(input_list) > 1, only
            # `input_list[0]` is passed to `evaluate`.
            # See https://github.com/pybamm-team/PyBaMM/pull/1261
            event.expression.evaluate(inputs=inputs_list[0])
            for event in model.discontinuity_events_eval
        ]

        # make sure they are increasing in time
        discontinuities = sorted(discontinuities)

        # remove any identical discontinuities
        discontinuities = [
            v
            for i, v in enumerate(discontinuities)
            if (
                i == len(discontinuities) - 1
                or discontinuities[i] < discontinuities[i + 1]
            )
            and v > 0
        ]

        # remove any discontinuities after end of t_eval
        discontinuities = [v for v in discontinuities if v < t_eval_dimensionless[-1]]

        if len(discontinuities) > 0:
            pybamm.logger.verbose(
                "Discontinuity events found at t = {}".format(discontinuities)
            )
            if isinstance(inputs, list):
                raise pybamm.SolverError(
                    "Cannot solve for a list of input parameters"
                    " sets with discontinuities"
                )
        else:
            pybamm.logger.verbose("No discontinuity events found")

        # insert time points around discontinuities in t_eval
        # keep track of sub sections to integrate by storing start and end indices
        start_indices = [0]
        end_indices = []
        eps = sys.float_info.epsilon
        for dtime in discontinuities:
            dindex = np.searchsorted(t_eval_dimensionless, dtime, side="left")
            end_indices.append(dindex + 1)
            start_indices.append(dindex + 1)
            if dtime - eps < t_eval_dimensionless[dindex] < dtime + eps:
                t_eval_dimensionless[dindex] += eps
                t_eval_dimensionless = np.insert(
                    t_eval_dimensionless, dindex, dtime - eps
                )
            else:
                t_eval_dimensionless = np.insert(
                    t_eval_dimensionless, dindex, [dtime - eps, dtime + eps]
                )
        end_indices.append(len(t_eval_dimensionless))

        # Integrate separately over each time segment and accumulate into the solution
        # object, restarting the solver at each discontinuity (and recalculating a
        # consistent state afterwards if a DAE)
        old_y0 = model.y0
        solutions = None
        for start_index, end_index in zip(start_indices, end_indices):
            pybamm.logger.verbose(
                "Calling solver for {} < t < {}".format(
                    t_eval_dimensionless[start_index] * model.timescale_eval,
                    t_eval_dimensionless[end_index - 1] * model.timescale_eval,
                )
            )
            ninputs = len(ext_and_inputs_list)
            if ninputs == 1:
                new_solution = self._integrate(
                    model,
                    t_eval_dimensionless[start_index:end_index],
                    ext_and_inputs_list[0],
                )
                new_solutions = [new_solution]
            else:
                with mp.Pool(processes=nproc) as p:
                    new_solutions = p.starmap(
                        self._integrate,
                        zip(
                            [model] * ninputs,
                            [t_eval_dimensionless[start_index:end_index]] * ninputs,
                            ext_and_inputs_list,
                        ),
                    )
                    p.close()
                    p.join()
            # Setting the solve time for each segment.
            # pybamm.Solution.__add__ assumes attribute solve_time.
            solve_time = timer.time()
            for sol in new_solutions:
                sol.solve_time = solve_time
            if start_index == start_indices[0]:
                solutions = [sol for sol in new_solutions]
            else:
                for i, new_solution in enumerate(new_solutions):
                    solutions[i] = solutions[i] + new_solution

            if solutions[0].termination != "final time":
                break

            if end_index != len(t_eval_dimensionless):
                # setup for next integration subsection
                last_state = solutions[0].y[:, -1]
                # update y0 (for DAE solvers, this updates the initial guess for the
                # rootfinder)
                model.y0 = last_state
                if len(model.algebraic) > 0:
                    model.y0 = self.calculate_consistent_state(
                        model, t_eval_dimensionless[end_index], ext_and_inputs_list[0]
                    )
        solve_time = timer.time()

        for i, solution in enumerate(solutions):
            # Check if extrapolation occurred
            extrapolation = self.check_extrapolation(solution, model.events)
            if extrapolation:
                warnings.warn(
                    "While solving {} extrapolation occurred for {}".format(
                        model.name, extrapolation
                    ),
                    pybamm.SolverWarning,
                )
            # Identify the event that caused termination and update the solution to
            # include the event time and state
            solutions[i], termination = self.get_termination_reason(
                solution, model.events
            )
            # Assign times
            solutions[i].set_up_time = set_up_time
            # all solutions get the same solve time, but their integration time
            # will be different (see https://github.com/pybamm-team/PyBaMM/pull/1261)
            solutions[i].solve_time = solve_time

        # Restore old y0
        model.y0 = old_y0

        # Report times
        if len(solutions) == 1:
            pybamm.logger.info("Finish solving {} ({})".format(model.name, termination))
            pybamm.logger.info(
                (
                    "Set-up time: {}, Solve time: {} (of which integration time: {}), "
                    "Total time: {}"
                ).format(
                    solutions[0].set_up_time,
                    solutions[0].solve_time,
                    solutions[0].integration_time,
                    solutions[0].total_time,
                )
            )
        else:
            pybamm.logger.info("Finish solving {} for all inputs".format(model.name))
            pybamm.logger.info(
                ("Set-up time: {}, Solve time: {}, Total time: {}").format(
                    solutions[0].set_up_time,
                    solutions[0].solve_time,
                    solutions[0].total_time,
                )
            )

        # Raise error if solutions[0] only contains one timestep (except for algebraic
        # solvers, where we may only expect one time in the solution)
        if (
            self.algebraic_solver is False
            and len(solution.all_ts) == 1
            and len(solution.all_ts[0]) == 1
        ):
            raise pybamm.SolverError(
                "Solution time vector has length 1. "
                "Check whether simulation terminated too early."
            )

        # Return solution(s)
        if ninputs == 1:
            return solutions[0]
        else:
            return solutions

    def step(
        self,
        old_solution,
        model,
        dt,
        npts=2,
        external_variables=None,
        inputs=None,
        save=True,
    ):
        """
        Step the solution of the model forward by a given time increment. The
        first time this method is called it executes the necessary setup by
        calling `self.set_up(model)`.

        Parameters
        ----------
        old_solution : :class:`pybamm.Solution` or None
            The previous solution to be added to. If `None`, a new solution is created.
        model : :class:`pybamm.BaseModel`
            The model whose solution to calculate. Must have attributes rhs and
            initial_conditions
        dt : numeric type
            The timestep (in seconds) over which to step the solution
        npts : int, optional
            The number of points at which the solution will be returned during
            the step dt. default is 2 (returns the solution at t0 and t0 + dt).
        external_variables : dict
            A dictionary of external variables and their corresponding
            values at the current time
        inputs_dict : dict, optional
            Any input parameters to pass to the model when solving
        save : bool
            Turn on to store the solution of all previous timesteps




        Raises
        ------
        :class:`pybamm.ModelError`
            If an empty model is passed (`model.rhs = {}` and `model.algebraic = {}` and
            `model.variables = {}`)

        """

        if old_solution is not None and not (
            old_solution.termination == "final time"
            or "[experiment]" in old_solution.termination
        ):
            # Return same solution as an event has already been triggered
            # With hack to allow stepping past experiment current / voltage cut-off
            return old_solution

        # Make sure model isn't empty
        if len(model.rhs) == 0 and len(model.algebraic) == 0:
            if not isinstance(self, pybamm.DummySolver):
                raise pybamm.ModelError(
                    "Cannot step empty model, use `pybamm.DummySolver` instead"
                )

        # Make sure dt is positive
        if dt <= 0:
            raise pybamm.SolverError("Step time must be positive")

        # Set timer
        timer = pybamm.Timer()

        # Set up external variables and inputs
        external_variables = external_variables or {}
        inputs = inputs or {}
        ext_and_inputs = {**external_variables, **inputs}

        # Check that any inputs that may affect the scaling have not changed
        # Set model timescale
        temp_timescale_eval = model.timescale.evaluate(inputs=inputs)
        # Set model lengthscales
        temp_length_scales_eval = {
            domain: scale.evaluate(inputs=inputs)
            for domain, scale in model.length_scales.items()
        }
        if old_solution is not None:
            if temp_timescale_eval != old_solution.timescale_eval:
                raise pybamm.SolverError(
                    "The model timescale is a function of an input parameter "
                    "and the value has changed between steps!"
                )
            for domain in temp_length_scales_eval.keys():
                old_dom_eval = old_solution.length_scales_eval[domain]
                if temp_length_scales_eval[domain] != old_dom_eval:
                    pybamm.logger.error(
                        "The {} domain lengthscale is a function of an input "
                        "parameter and the value has changed between "
                        "steps!".format(domain)
                    )

        if old_solution is None:
            # Run set up on first step
            pybamm.logger.verbose(
                "Start stepping {} with {}".format(model.name, self.name)
            )

            self.set_up(model, ext_and_inputs)
            self.models_set_up.update(
                {model: {"initial conditions": model.concatenated_initial_conditions}}
            )
            t = 0.0
        elif model not in self.models_set_up:
            # Run set up if the model has changed

            self.set_up(model, ext_and_inputs)
            self.models_set_up.update(
                {model: {"initial conditions": model.concatenated_initial_conditions}}
            )

        if old_solution is not None:
            t = old_solution.all_ts[-1][-1]
            if old_solution.all_models[-1] == model:
                # initialize with old solution
                model.y0 = old_solution.all_ys[-1][:, -1]
            else:
                model.y0 = (
                    model.set_initial_conditions_from(old_solution)
                    .concatenated_initial_conditions.evaluate(0, inputs=ext_and_inputs)
                    .flatten()
                )
        set_up_time = timer.time()

        # (Re-)calculate consistent initial conditions
        self._set_initial_conditions(model, ext_and_inputs, update_rhs=False)

        # Non-dimensionalise dt
        dt_dimensionless = dt / model.timescale_eval

        # Step
        t_eval = np.linspace(t, t + dt_dimensionless, npts)
        pybamm.logger.verbose(
            "Stepping for {:.0f} < t < {:.0f}".format(
                t * model.timescale_eval,
                (t + dt_dimensionless) * model.timescale_eval,
            )
        )
        timer.reset()
        solution = self._integrate(model, t_eval, ext_and_inputs)
        solution.solve_time = timer.time()

        # Check if extrapolation occurred
        extrapolation = self.check_extrapolation(solution, model.events)
        if extrapolation:
            warnings.warn(
                "While solving {} extrapolation occurred for {}".format(
                    model.name, extrapolation
                ),
                pybamm.SolverWarning,
            )

        # Identify the event that caused termination and update the solution to
        # include the event time and state
        solution, termination = self.get_termination_reason(solution, model.events)

        # Assign setup time
        solution.set_up_time = set_up_time

        # Report times
        pybamm.logger.verbose("Finish stepping {} ({})".format(model.name, termination))
        pybamm.logger.verbose(
            (
                "Set-up time: {}, Step time: {} (of which integration time: {}), "
                "Total time: {}"
            ).format(
                solution.set_up_time,
                solution.solve_time,
                solution.integration_time,
                solution.total_time,
            )
        )

        # Return solution
        if save is False:
            return solution
        else:
            return old_solution + solution

    def get_termination_reason(self, solution, events):
        """
        Identify the cause for termination. In particular, if the solver terminated
        due to an event, (try to) pinpoint which event was responsible. If an event
        occurs the event time and state are added to the solution object.
        Note that the current approach (evaluating all the events and then finding which
        one is smallest at the final timestep) is pretty crude, but is the easiest one
        that works for all the different solvers.

        Parameters
        ----------
        solution : :class:`pybamm.Solution`
            The solution object
        events : dict
            Dictionary of events
        """
        termination_events = [
            x for x in events if x.event_type == pybamm.EventType.TERMINATION
        ]
        if solution.termination == "final time":
            return (
                solution,
                "the solver successfully reached the end of the integration interval",
            )
        elif solution.termination == "event":
            pybamm.logger.debug("Start post-processing events")
            if solution.closest_event_idx is not None:
                solution.termination = (
                    f"event: {termination_events[solution.closest_event_idx].name}"
                )
            else:
                # Get final event value
                final_event_values = {}

                for event in termination_events:
                    final_event_values[event.name] = abs(
                        event.expression.evaluate(
                            solution.t_event,
                            solution.y_event,
                            inputs=solution.all_inputs[-1],
                        )
                    )
                termination_event = min(final_event_values, key=final_event_values.get)

                # Check that it's actually an event
                if abs(final_event_values[termination_event]) > 0.1:  # pragma: no cover
                    # Hard to test this
                    raise pybamm.SolverError(
                        "Could not determine which event was triggered "
                        "(possibly due to NaNs)"
                    )
                # Add the event to the solution object
                solution.termination = "event: {}".format(termination_event)
            # Update t, y and inputs to include event time and state
            # Note: if the final entry of t is equal to the event time we skip
            # this (having duplicate entries causes an error later in ProcessedVariable)
            if solution.t_event != solution.all_ts[-1][-1]:
                event_sol = pybamm.Solution(
                    solution.t_event,
                    solution.y_event,
                    solution.all_models[-1],
                    solution.all_inputs[-1],
                    solution.t_event,
                    solution.y_event,
                    solution.termination,
                )
                event_sol.solve_time = 0
                event_sol.integration_time = 0
                solution = solution + event_sol

            pybamm.logger.debug("Finish post-processing events")
            return solution, solution.termination
        elif solution.termination == "success":
            return solution, solution.termination

    def check_extrapolation(self, solution, events):
        """
        Check if extrapolation occurred for any of the interpolants. Note that with the
        current approach (evaluating all the events at the solution times) some
        extrapolations might not be found if they only occurred for a small period of
        time.

        Parameters
        ----------
        solution : :class:`pybamm.Solution`
            The solution object
        events : dict
            Dictionary of events
        """
        extrap_events = {}

        for event in events:
            if event.event_type == pybamm.EventType.INTERPOLANT_EXTRAPOLATION:
                # First set to False, then loop through and change to True if any
                # events extrapolate
                extrap_events[event.name] = False
                # This might be a little bit slow but is ok for now
                for ts, ys, inputs in zip(
                    solution.all_ts, solution.all_ys, solution.all_inputs
                ):
                    for inner_idx, t in enumerate(ts):
                        y = ys[:, inner_idx]
                        if isinstance(y, casadi.DM):
                            y = y.full()
                        if (
                            event.expression.evaluate(t, y, inputs=inputs)
                            < self.extrap_tol
                        ):
                            extrap_events[event.name] = True

        # Add the event dictionaryto the solution object
        solution.extrap_events = extrap_events

        return [k for k, v in extrap_events.items() if v]

    def _set_up_ext_and_inputs(
        self, model, external_variables, inputs, calculate_sensitivities
    ):
        """Set up external variables and input parameters"""
        inputs = inputs or {}

        # Go through all input parameters that can be found in the model
        # If any of them are *not* provided by "inputs", a symbolic input parameter is
        # created, with appropriate size
        for input_param in model.input_parameters:
            name = input_param.name
            if name not in inputs:
                # Only allow symbolic inputs for CasadiSolver and CasadiAlgebraicSolver
                if not isinstance(
                    self, (pybamm.CasadiSolver, pybamm.CasadiAlgebraicSolver)
                ):
                    raise pybamm.SolverError(
                        "Only CasadiSolver and CasadiAlgebraicSolver "
                        "can have symbolic inputs"
                    )
                inputs[name] = casadi.MX.sym(name, input_param._expected_size)

        external_variables = external_variables or {}
        ext_and_inputs = {**external_variables, **inputs}
        return ext_and_inputs


class SolverCallable:
    """A class that will be called by the solver when integrating"""

    def __init__(self, function, name, model):
        self._function = function
        if isinstance(function, casadi.Function):
            self.form = "casadi"
        else:
            self.form = "python"
        self.name = name
        self.model = model
        self.timescale = self.model.timescale_eval

    def __call__(self, t, y, inputs):
        pybamm.logger.debug(
            "Evaluating {} for {} at t={}".format(
                self.name, self.model.name, t * self.timescale
            )
        )
        if self.name in ["RHS", "algebraic", "residuals", "event"]:

            return self.function(t, y, inputs).flatten()
        else:
            return self.function(t, y, inputs)

    def function(self, t, y, inputs):
        if self.form == "casadi":
            states_eval = self._function(t, y, inputs)
            if self.name in ["RHS", "algebraic", "residuals", "event"]:
                return states_eval.full()
            else:
                # keep jacobians sparse
                return states_eval
        else:
            return self._function(t, y, inputs=inputs, known_evals={})[0]


class SensitivityCallable:
    """A class that will be called by the solver when integrating"""

    def __init__(self, function, name, model, form):
        self._function = function
        self.form = form
        self.name = name
        self.model = model
        self.timescale = self.model.timescale_eval

    def __call__(self, t, y, inputs):
        pybamm.logger.debug(
            "Evaluating sensitivities of {} for {} at t={}".format(
                self.name, self.model.name, t * self.timescale
            )
        )
        return self.function(t, y, inputs)

    def function(self, t, y, inputs):
        if self.form == "casadi":
            return self._function(t, y, inputs)
        elif self.form == "jax":
            return self._function(t, y, inputs=inputs)
        else:
            ret_with_known_evals = self._function(t, y, inputs=inputs, known_evals={})
            return {k: v[0] for k, v in ret_with_known_evals.items()}


class Residuals(SolverCallable):
    """Returns information about residuals at time t and state y"""

    def __init__(self, function, name, model):
        super().__init__(function, name, model)
        if model.mass_matrix is not None:
            self.mass_matrix = model.mass_matrix.entries

    def __call__(self, t, y, ydot, inputs):
        states_eval = super().__call__(t, y, inputs)
        return states_eval - self.mass_matrix @ ydot


class InitialConditions(SolverCallable):
    """Returns initial conditions given inputs"""

    def __init__(self, function, model):
        super().__init__(function, "initial conditions", model)

    def __call__(self, inputs):
        if self.form == "casadi":
            if isinstance(inputs, dict):
                inputs = casadi.vertcat(*[x for x in inputs.values()])
            return self._function(0, self.y_dummy, inputs)
        else:
            return self._function(0, self.y_dummy, inputs=inputs).flatten()



#
# Settings class for PyBaMM
#


class Settings(object):
    _debug_mode = False
    _simplify = True
    _min_smoothing = "exact"
    _max_smoothing = "exact"
    _heaviside_smoothing = "exact"
    _abs_smoothing = "exact"
    max_words_in_line = 4

    @property
    def debug_mode(self):
        return self._debug_mode

    @debug_mode.setter
    def debug_mode(self, value):
        assert isinstance(value, bool)
        self._debug_mode = value

    @property
    def simplify(self):
        return self._simplify

    @simplify.setter
    def simplify(self, value):
        assert isinstance(value, bool)
        self._simplify = value

    def set_smoothing_parameters(self, k):
        "Helper function to set all smoothing parameters"
        self.min_smoothing = k
        self.max_smoothing = k
        self.heaviside_smoothing = k
        self.abs_smoothing = k

    @property
    def min_smoothing(self):
        return self._min_smoothing

    @min_smoothing.setter
    def min_smoothing(self, k):
        if k != "exact" and k <= 0:
            raise ValueError(
                "smoothing parameter must be 'exact' or a strictly positive number"
            )
        self._min_smoothing = k

    @property
    def max_smoothing(self):
        return self._max_smoothing

    @max_smoothing.setter
    def max_smoothing(self, k):
        if k != "exact" and k <= 0:
            raise ValueError(
                "smoothing parameter must be 'exact' or a strictly positive number"
            )
        self._max_smoothing = k

    @property
    def heaviside_smoothing(self):
        return self._heaviside_smoothing

    @heaviside_smoothing.setter
    def heaviside_smoothing(self, k):
        if k != "exact" and k <= 0:
            raise ValueError(
                "smoothing parameter must be 'exact' or a strictly positive number"
            )
        self._heaviside_smoothing = k

    @property
    def abs_smoothing(self):
        return self._abs_smoothing

    @abs_smoothing.setter
    def abs_smoothing(self, k):
        if k != "exact" and k <= 0:
            raise ValueError(
                "smoothing parameter must be 'exact' or a strictly positive number"
            )
        self._abs_smoothing = k


settings = Settings()



import pybamm
from .dfn import DFN


class Yang2017(DFN):
    def __init__(self, options=None, name="Yang2017", build=True):
        options = {
            "SEI": "ec reaction limited",
            "SEI film resistance": "distributed",
            "SEI porosity change": "true",
            "lithium plating": "irreversible",
            "lithium plating porosity change": "true",
        }
        super().__init__(options=options, name=name)
        pybamm.citations.register("Yang2017")

    @property
    def default_parameter_values(self):
        return pybamm.ParameterValues(chemistry=pybamm.parameter_sets.Chen2020_plating)



#
# Class for a particle-size distribution averaged in the x direction,
# with Fickian diffusion within each particle
#
import pybamm

from .base_distribution import BaseSizeDistribution


class XAveragedFickianDiffusion(BaseSizeDistribution):
    """Class for molar conservation in an x-averaged particle-size
    distribution with Fickian diffusion in each particle. Concentration
    varies with r (spherical coordinate), R (particle size) but not x.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    domain : str
        The domain of the model either 'Negative' or 'Positive'


    **Extends:** :class:`pybamm.particle.size_distribution.BaseSizeDistribution`
    """

    def __init__(self, param, domain):
        super().__init__(param, domain)
        pybamm.citations.register("Kirk2020")

    def get_fundamental_variables(self):
        if self.domain == "Negative":
            # distribution variables
            c_s_xav_distribution = pybamm.Variable(
                "X-averaged negative particle concentration distribution",
                domain="negative particle",
                auxiliary_domains={
                    "secondary": "negative particle size",
                    "tertiary": "current collector",
                },
                bounds=(0, 1),
            )
            # Since concentration does not depend on "x", need a particle-size
            # spatial variable R with only "current collector" as an auxiliary
            # domain
            R = pybamm.SpatialVariable(
                "R_n",
                domain=["negative particle size"],
                auxiliary_domains={"secondary": "current collector"},
                coord_sys="cartesian",
            )

        elif self.domain == "Positive":
            # distribution variables
            c_s_xav_distribution = pybamm.Variable(
                "X-averaged positive particle concentration distribution",
                domain="positive particle",
                auxiliary_domains={
                    "secondary": "positive particle size",
                    "tertiary": "current collector",
                },
                bounds=(0, 1),
            )
            # Since concentration does not depend on "x", need a particle-size
            # spatial variable R with only "current collector" as an auxiliary
            # domain
            R = pybamm.SpatialVariable(
                "R_p",
                domain=["positive particle size"],
                auxiliary_domains={"secondary": "current collector"},
                coord_sys="cartesian",
            )

        # Distribution variables
        variables = self._get_distribution_variables(R)

        # Concentration distribution variables (R-dependent)
        variables.update(
            self._get_standard_concentration_distribution_variables(
                c_s_xav_distribution
            )
        )

        # Standard R-averaged variables. Average concentrations using
        # the volume-weighted distribution since they are volume-based
        # quantities. Necessary for output variables "Total lithium in
        # negative electrode [mol]", etc, to be calculated correctly
        f_v_dist = variables[
            "X-averaged "
            + self.domain.lower()
            + " volume-weighted particle-size distribution"
        ]
        c_s_xav = pybamm.Integral(f_v_dist * c_s_xav_distribution, R)
        c_s = pybamm.SecondaryBroadcast(c_s_xav, [self.domain.lower() + " electrode"])
        variables.update(self._get_standard_concentration_variables(c_s, c_s_xav))
        return variables

    def get_coupled_variables(self, variables):
        c_s_xav_distribution = variables[
            "X-averaged " + self.domain.lower() + " particle concentration distribution"
        ]
        R = variables[self.domain + " particle sizes"]

        # broadcast to "particle size" domain then again into "particle"
        T_k_xav = pybamm.PrimaryBroadcast(
            variables["X-averaged " + self.domain.lower() + " electrode temperature"],
            [self.domain.lower() + " particle size"],
        )
        T_k_xav = pybamm.PrimaryBroadcast(
            T_k_xav,
            [self.domain.lower() + " particle"],
        )

        if self.domain == "Negative":
            N_s_xav_distribution = -self.param.D_n(
                c_s_xav_distribution, T_k_xav
            ) * pybamm.grad(c_s_xav_distribution)
        elif self.domain == "Positive":
            N_s_xav_distribution = -self.param.D_p(
                c_s_xav_distribution, T_k_xav
            ) * pybamm.grad(c_s_xav_distribution)

        # Size-dependent flux variables
        variables.update(
            self._get_standard_flux_distribution_variables(N_s_xav_distribution)
        )

        # Size-averaged flux variables (perform area-weighted avg manually as flux
        # evals on edges)
        f_a_dist = pybamm.PrimaryBroadcast(
            variables[
                "X-averaged "
                + self.domain.lower()
                + " area-weighted particle-size distribution"
            ],
            [self.domain.lower() + " particle"],
        )
        N_s_xav = pybamm.Integral(f_a_dist * N_s_xav_distribution, R)
        N_s = pybamm.SecondaryBroadcast(N_s_xav, [self.domain.lower() + " electrode"])
        variables.update(self._get_standard_flux_variables(N_s, N_s_xav))

        variables.update(self._get_total_concentration_variables(variables))
        return variables

    def set_rhs(self, variables):
        # Extract x-av variables
        c_s_xav_distribution = variables[
            "X-averaged " + self.domain.lower() + " particle concentration distribution"
        ]

        N_s_xav_distribution = variables[
            "X-averaged " + self.domain.lower() + " particle flux distribution"
        ]

        # Spatial variable R, broadcast into particle
        R_spatial_variable = variables[self.domain + " particle sizes"]
        R = pybamm.PrimaryBroadcast(
            R_spatial_variable,
            [self.domain.lower() + " particle"],
        )
        if self.domain == "Negative":
            self.rhs = {
                c_s_xav_distribution: -(1 / self.param.C_n)
                * pybamm.div(N_s_xav_distribution)
                / R ** 2
            }
        elif self.domain == "Positive":
            self.rhs = {
                c_s_xav_distribution: -(1 / self.param.C_p)
                * pybamm.div(N_s_xav_distribution)
                / R ** 2
            }

    def set_boundary_conditions(self, variables):
        # Extract x-av variables
        c_s_xav_distribution = variables[
            "X-averaged " + self.domain.lower() + " particle concentration distribution"
        ]
        c_s_surf_xav_distribution = variables[
            "X-averaged "
            + self.domain.lower()
            + " particle surface concentration distribution"
        ]
        j_xav_distribution = variables[
            "X-averaged "
            + self.domain.lower()
            + " electrode interfacial current density distribution"
        ]
        R = variables[self.domain + " particle sizes"]

        # Extract x-av T and broadcast to particle size domain
        T_k_xav = variables[
            "X-averaged " + self.domain.lower() + " electrode temperature"
        ]
        T_k_xav = pybamm.PrimaryBroadcast(
            T_k_xav, [self.domain.lower() + " particle size"]
        )

        # Set surface Neumann boundary values
        if self.domain == "Negative":
            rbc = (
                -self.param.C_n
                * R
                * j_xav_distribution
                / self.param.a_R_n
                / self.param.D_n(c_s_surf_xav_distribution, T_k_xav)
            )

        elif self.domain == "Positive":
            rbc = (
                -self.param.C_p
                * R
                * j_xav_distribution
                / self.param.a_R_p
                / self.param.gamma_p
                / self.param.D_p(c_s_surf_xav_distribution, T_k_xav)
            )

        self.boundary_conditions = {
            c_s_xav_distribution: {
                "left": (pybamm.Scalar(0), "Neumann"),
                "right": (rbc, "Neumann"),
            }
        }

    def set_initial_conditions(self, variables):
        """
        For x-averaged particle-size distribution models, initial conditions can't
        depend on x so we arbitrarily set the initial values of the single
        particles to be given by the values at x=0 in the negative electrode
        and x=1 in the positive electrode. Typically, supplied initial
        conditions are uniform x.
        """
        c_s_xav_distribution = variables[
            "X-averaged " + self.domain.lower() + " particle concentration distribution"
        ]

        if self.domain == "Negative":
            c_init = self.param.c_n_init(0)

        elif self.domain == "Positive":
            c_init = self.param.c_p_init(1)

        self.initial_conditions = {c_s_xav_distribution: c_init}



#
# Solver class using sundials with the KLU sparse linear solver
#
import casadi
import pybamm
import numpy as np
import scipy.sparse as sparse

import importlib

idaklu_spec = importlib.util.find_spec("pybamm.solvers.idaklu")
if idaklu_spec is not None:
    try:
        idaklu = importlib.util.module_from_spec(idaklu_spec)
        idaklu_spec.loader.exec_module(idaklu)
    except ImportError:  # pragma: no cover
        idaklu_spec = None


def have_idaklu():
    return idaklu_spec is not None


class IDAKLUSolver(pybamm.BaseSolver):
    """Solve a discretised model, using sundials with the KLU sparse linear solver.

     Parameters
    ----------
    rtol : float, optional
        The relative tolerance for the solver (default is 1e-6).
    atol : float, optional
        The absolute tolerance for the solver (default is 1e-6).
    root_method : str or pybamm algebraic solver class, optional
        The method to use to find initial conditions (for DAE solvers).
        If a solver class, must be an algebraic solver class.
        If "casadi",
        the solver uses casadi's Newton rootfinding algorithm to find initial
        conditions. Otherwise, the solver uses 'scipy.optimize.root' with method
        specified by 'root_method' (e.g. "lm", "hybr", ...)
    root_tol : float, optional
        The tolerance for the initial-condition solver (default is 1e-6).
    extrap_tol : float, optional
        The tolerance to assert whether extrapolation occurs or not (default is 0).
    """

    def __init__(
        self,
        rtol=1e-6,
        atol=1e-6,
        root_method="casadi",
        root_tol=1e-6,
        extrap_tol=0,
        max_steps="deprecated",
    ):

        if idaklu_spec is None:  # pragma: no cover
            raise ImportError("KLU is not installed")

        super().__init__(
            "ida",
            rtol,
            atol,
            root_method,
            root_tol,
            extrap_tol,
            max_steps,
        )
        self.name = "IDA KLU solver"

        pybamm.citations.register("Hindmarsh2000")
        pybamm.citations.register("Hindmarsh2005")

    def set_atol_by_variable(self, variables_with_tols, model):
        """
        A method to set the absolute tolerances in the solver by state variable.
        This method attaches a vector of tolerance to the model. (i.e. model.atol)

        Parameters
        ----------
        variables_with_tols : dict
            A dictionary with keys that are strings indicating the variable you
            wish to set the tolerance of and values that are the tolerances.

        model : :class:`pybamm.BaseModel`
            The model that is going to be solved.
        """

        size = model.concatenated_initial_conditions.size
        atol = self._check_atol_type(self.atol, size)
        for var, tol in variables_with_tols.items():
            variable = model.variables[var]
            if isinstance(variable, pybamm.StateVector):
                atol = self.set_state_vec_tol(atol, variable, tol)
            else:
                raise pybamm.SolverError("Can only set tolerances for state variables")

        model.atol = atol

    def set_state_vec_tol(self, atol, state_vec, tol):
        """
        A method to set the tolerances in the atol vector of a specific
        state variable. This method modifies self.atol

        Parameters
        ----------
        state_vec : :class:`pybamm.StateVector`
            The state vector to apply to the tolerance to
        tol: float
            The tolerance value
        """
        slices = state_vec.y_slices[0]
        atol[slices] = tol
        return atol

    def _check_atol_type(self, atol, size):
        """
        This method checks that the atol vector is of the right shape and
        type.

        Parameters
        ----------
        atol: double or np.array or list
            Absolute tolerances. If this is a vector then each entry corresponds to
            the absolute tolerance of one entry in the state vector.
        size: int
            The length of the atol vector
        """

        if isinstance(atol, float):
            atol = atol * np.ones(size)
        elif isinstance(atol, list):
            atol = np.array(atol)
        elif isinstance(atol, np.ndarray):
            pass
        else:
            raise pybamm.SolverError(
                "Absolute tolerances must be a numpy array, float, or list"
            )

        if atol.size != size:
            raise pybamm.SolverError(
                """Absolute tolerances must be either a scalar or a numpy arrray
                of the same shape as y0 ({})""".format(
                    size
                )
            )

        return atol

    def _integrate(self, model, t_eval, inputs_dict=None):
        """
        Solve a DAE model defined by residuals with initial conditions y0.

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The model whose solution to calculate.
        t_eval : numeric type
            The times at which to compute the solution
        inputs_dict : dict, optional
            Any external variables or input parameters to pass to the model when solving
        """
        inputs_dict = inputs_dict or {}
        if model.rhs_eval.form == "casadi":
            # stack inputs
            inputs = casadi.vertcat(*[x for x in inputs_dict.values()])
            # raise warning about casadi format being slow
            pybamm.logger.warning(
                "Using casadi form for the IDA KLU solver is slow. "
                "Set `model.convert_to_format='python'` for better performance. "
                "For DAE models, this may also require changing the root method to "
                "'lm'."
            )
        else:
            inputs = inputs_dict

        if model.jacobian_eval is None:
            raise pybamm.SolverError("KLU requires the Jacobian to be provided")

        try:
            atol = model.atol
        except AttributeError:
            atol = self.atol

        y0 = model.y0
        if isinstance(y0, casadi.DM):
            y0 = y0.full().flatten()

        rtol = self.rtol
        atol = self._check_atol_type(atol, y0.size)

        if model.convert_to_format == "jax":
            mass_matrix = model.mass_matrix.entries.toarray()
        else:
            mass_matrix = model.mass_matrix.entries

        # construct residuals function by binding inputs
        def resfn(t, y, ydot):
            return model.residuals_eval(t, y, ydot, inputs)

        if model.jacobian_eval:
            jac_y0_t0 = model.jacobian_eval(t_eval[0], y0, inputs)
            if sparse.issparse(jac_y0_t0):

                def jacfn(t, y, cj):
                    j = model.jacobian_eval(t, y, inputs) - cj * mass_matrix
                    return j

            else:

                def jacfn(t, y, cj):
                    jac_eval = model.jacobian_eval(t, y, inputs) - cj * mass_matrix
                    return sparse.csr_matrix(jac_eval)

        class SundialsJacobian:
            def __init__(self):
                self.J = None

                random = np.random.random(size=y0.size)
                J = jacfn(10, random, 20)
                self.nnz = J.nnz  # hoping nnz remains constant...

            def jac_res(self, t, y, cj):
                # must be of form j_res = (dr/dy) - (cj) (dr/dy')
                # cj is just the input parameter
                # see p68 of the ida_guide.pdf for more details
                self.J = jacfn(t, y, cj)

            def get_jac_data(self):
                return self.J.data

            def get_jac_row_vals(self):
                return self.J.indices

            def get_jac_col_ptrs(self):
                return self.J.indptr

        # solver works with ydot0 set to zero
        ydot0 = np.zeros_like(y0)

        jac_class = SundialsJacobian()

        num_of_events = len(model.terminate_events_eval)
        use_jac = 1

        def rootfn(t, y):
            return_root = np.ones((num_of_events,))
            return_root[:] = [
                event(t, y, inputs) for event in model.terminate_events_eval
            ]

            return return_root

        # get ids of rhs and algebraic variables
        rhs_ids = np.ones(model.rhs_eval(0, y0, inputs).shape[0])
        alg_ids = np.zeros(len(y0) - len(rhs_ids))
        ids = np.concatenate((rhs_ids, alg_ids))

        number_of_sensitivity_parameters = 0
        if model.sensitivities_eval is not None:
            sens0 = model.sensitivities_eval(t=0, y=y0, inputs=inputs)
            number_of_sensitivity_parameters = len(sens0.keys())

        def sensfn(resvalS, t, y, yp, yS, ypS):
            """
            this function evaluates the sensitivity equations required by IDAS,
            returning them in resvalS, which is preallocated as a numpy array of size
            (np, n), where n is the number of states and np is the number of parameters

            The equations returned are:

             dF/dy * s_i + dF/dyd * sd_i + dFdp_i for i in range(np)

            Parameters
            ----------
            resvalS: ndarray of shape (np, n)
                returns the sensitivity equations in this preallocated array
            t: number
                time value
            y: ndarray of shape (n)
                current state vector
            yp: list (np) of ndarray of shape (n)
                current time derivative of state vector
            yS: list (np) of ndarray of shape (n)
                current state vector of sensitivity equations
            ypS: list (np) of ndarray of shape (n)
                current time derivative of state vector of sensitivity equations

            """

            dFdy = model.jacobian_eval(t, y, inputs)
            dFdyd = mass_matrix
            dFdp = model.sensitivities_eval(t, y, inputs)

            for i, dFdp_i in enumerate(dFdp.values()):
                resvalS[i][:] = dFdy @ yS[i] - dFdyd @ ypS[i] + dFdp_i

        # solve
        timer = pybamm.Timer()
        sol = idaklu.solve(
            t_eval,
            y0,
            ydot0,
            resfn,
            jac_class.jac_res,
            sensfn,
            jac_class.get_jac_data,
            jac_class.get_jac_row_vals,
            jac_class.get_jac_col_ptrs,
            jac_class.nnz,
            rootfn,
            num_of_events,
            use_jac,
            ids,
            atol,
            rtol,
            number_of_sensitivity_parameters,
        )
        integration_time = timer.time()

        t = sol.t
        number_of_timesteps = t.size
        number_of_states = y0.size
        y_out = sol.y.reshape((number_of_timesteps, number_of_states))

        # return sensitivity solution, we need to flatten yS to
        # (#timesteps * #states,) to match format used by Solution
        if number_of_sensitivity_parameters != 0:
            yS_out = {
                name: sol.yS[i].reshape(-1, 1) for i, name in enumerate(sens0.keys())
            }
        else:
            yS_out = False
        if sol.flag in [0, 2]:
            # 0 = solved for all t_eval
            if sol.flag == 0:
                termination = "final time"
            # 2 = found root(s)
            elif sol.flag == 2:
                termination = "event"

            sol = pybamm.Solution(
                sol.t,
                np.transpose(y_out),
                model,
                inputs_dict,
                t[-1],
                np.transpose(y_out[-1])[:, np.newaxis],
                termination,
                sensitivities=yS_out,
            )
            sol.integration_time = integration_time
            return sol
        else:
            raise pybamm.SolverError(sol.message)



#
# Unary operator classes and methods
#
import numbers

import numpy as np
from scipy.sparse import csr_matrix

import pybamm


class Broadcast(pybamm.SpatialOperator):
    """
    A node in the expression tree representing a broadcasting operator.
    Broadcasts a child to a specified domain. After discretisation, this will evaluate
    to an array of the right shape for the specified domain.

    For an example of broadcasts in action, see
    `this example notebook
    <https://github.com/pybamm-team/PyBaMM/blob/develop/examples/notebooks/expression_tree/broadcasts.ipynb>`_

    Parameters
    ----------
    child : :class:`Symbol`
        child node
    broadcast_domain : iterable of str
        Primary domain for broadcast. This will become the domain of the symbol
    broadcast_auxiliary_domains : dict of str
        Auxiliary domains for broadcast.
    broadcast_type : str, optional
        Whether to broadcast to the full domain (primary and secondary) or only in the
        primary direction. Default is "full".
    name : str
        name of the node

    **Extends:** :class:`SpatialOperator`
    """

    def __init__(
        self,
        child,
        broadcast_domain,
        broadcast_auxiliary_domains=None,
        broadcast_type="full to nodes",
        name=None,
    ):
        # Convert child to scalar if it is a number
        if isinstance(child, numbers.Number):
            child = pybamm.Scalar(child)
        # Convert domain to list if it's a string
        if isinstance(broadcast_domain, str):
            broadcast_domain = [broadcast_domain]

        if name is None:
            name = "broadcast"

        # perform some basic checks and set attributes
        domain, auxiliary_domains = self.check_and_set_domains(
            child, broadcast_type, broadcast_domain, broadcast_auxiliary_domains
        )
        self.broadcast_type = broadcast_type
        self.broadcast_domain = broadcast_domain
        super().__init__(name, child, domain, auxiliary_domains)

    @property
    def broadcasts_to_nodes(self):
        if self.broadcast_type.endswith("nodes"):
            return True
        else:
            return False

    def _sympy_operator(self, child):
        """Override :meth:`pybamm.UnaryOperator._sympy_operator`"""
        return child


class PrimaryBroadcast(Broadcast):
    """
    A node in the expression tree representing a primary broadcasting operator.
    Broadcasts in a `primary` dimension only. That is, makes explicit copies of the
    symbol in the domain specified by `broadcast_domain`. This should be used for
    broadcasting from a "larger" scale to a "smaller" scale, for example broadcasting
    temperature T(x) from the electrode to the particles, or broadcasting current
    collector current i(y, z) from the current collector to the electrodes.

    Parameters
    ----------
    child : :class:`Symbol`
        child node
    broadcast_domain : iterable of str
        Primary domain for broadcast. This will become the domain of the symbol
    name : str
        name of the node

    **Extends:** :class:`Broadcast`
    """

    def __init__(self, child, broadcast_domain, name=None):
        super().__init__(
            child, broadcast_domain, broadcast_type="primary to nodes", name=name
        )

    def check_and_set_domains(
        self, child, broadcast_type, broadcast_domain, broadcast_auxiliary_domains
    ):
        """See :meth:`Broadcast.check_and_set_domains`"""
        # Can only do primary broadcast from current collector to electrode,
        # particle-size or particle or from electrode to particle-size or particle.
        # Note e.g. current collector to particle *is* allowed
        if child.domain == []:
            pass
        elif child.domain == ["current collector"] and broadcast_domain[0] not in [
            "negative electrode",
            "separator",
            "positive electrode",
            "negative particle size",
            "positive particle size",
            "negative particle",
            "positive particle",
        ]:
            raise pybamm.DomainError(
                """Primary broadcast from current collector domain must be to electrode
                or separator or particle or particle size domains"""
            )
        elif (
            child.domain[0]
            in [
                "negative electrode",
                "separator",
                "positive electrode",
            ]
            and broadcast_domain[0] not in [
                "negative particle",
                "positive particle",
                "negative particle size",
                "positive particle size",
            ]
        ):
            raise pybamm.DomainError(
                """Primary broadcast from electrode or separator must be to particle
                or particle size domains"""
            )
        elif child.domain[0] in [
            "negative particle size",
            "positive particle size",
        ] and broadcast_domain[0] not in ["negative particle", "positive particle"]:
            raise pybamm.DomainError(
                """Primary broadcast from particle size domain must be to particle
                domain"""
            )
        elif child.domain[0] in ["negative particle", "positive particle"]:
            raise pybamm.DomainError("Cannot do primary broadcast from particle domain")

        domain = broadcast_domain
        auxiliary_domains = {}
        if child.domain != []:
            auxiliary_domains["secondary"] = child.domain
        if "secondary" in child.auxiliary_domains:
            auxiliary_domains["tertiary"] = child.auxiliary_domains["secondary"]
            if "tertiary" in child.auxiliary_domains:
                auxiliary_domains["quaternary"] = child.auxiliary_domains["tertiary"]

        return domain, auxiliary_domains

    def _unary_new_copy(self, child):
        """See :meth:`pybamm.UnaryOperator._unary_new_copy()`."""
        return self.__class__(child, self.broadcast_domain)

    def _evaluate_for_shape(self):
        """
        Returns a vector of NaNs to represent the shape of a Broadcast.
        See :meth:`pybamm.Symbol.evaluate_for_shape_using_domain()`
        """
        child_eval = self.children[0].evaluate_for_shape()
        vec = pybamm.evaluate_for_shape_using_domain(self.domain)
        return np.outer(child_eval, vec).reshape(-1, 1)

    def reduce_one_dimension(self):
        """Reduce the broadcast by one dimension."""
        return self.orphans[0]


class PrimaryBroadcastToEdges(PrimaryBroadcast):
    """A primary broadcast onto the edges of the domain."""

    def __init__(self, child, broadcast_domain, name=None):
        name = name or "broadcast to edges"
        super().__init__(child, broadcast_domain, name)
        self.broadcast_type = "primary to edges"

    def _evaluates_on_edges(self, dimension):
        return True


class SecondaryBroadcast(Broadcast):
    """
    A node in the expression tree representing a secondary broadcasting operator.
    Broadcasts in a `secondary` dimension only. That is, makes explicit copies of the
    symbol in the domain specified by `broadcast_domain`. This should be used for
    broadcasting from a "smaller" scale to a "larger" scale, for example broadcasting
    SPM particle concentrations c_s(r) from the particles to the electrodes. Note that
    this wouldn't be used to broadcast particle concentrations in the DFN, since these
    already depend on both x and r.

    Parameters
    ----------
    child : :class:`Symbol`
        child node
    broadcast_domain : iterable of str
        Secondary domain for broadcast. This will become the secondary domain of the
        symbol, shifting the child's `secondary` and `tertiary` (if present) over by
        one position.
    name : str
        name of the node

    **Extends:** :class:`Broadcast`
    """

    def __init__(self, child, broadcast_domain, name=None):
        super().__init__(
            child, broadcast_domain, broadcast_type="secondary to nodes", name=name
        )

    def check_and_set_domains(
        self, child, broadcast_type, broadcast_domain, broadcast_auxiliary_domains
    ):
        """See :meth:`Broadcast.check_and_set_domains`"""
        if child.domain == []:
            raise TypeError(
                "Cannot take SecondaryBroadcast of an object with empty domain. "
                "Use PrimaryBroadcast instead."
            )
        # Can only do secondary broadcast from particle to electrode or from
        # electrode to current collector
        if child.domain[0] in [
            "negative particle",
            "positive particle",
        ] and broadcast_domain[0] not in [
            "negative particle size",
            "positive particle size",
            "negative electrode",
            "separator",
            "positive electrode",
        ]:
            raise pybamm.DomainError(
                """Secondary broadcast from particle domain must be to particle-size,
                electrode or separator domains"""
            )
        if child.domain[0] in [
            "negative particle size",
            "positive particle size",
        ] and broadcast_domain[0] not in [
            "negative electrode",
            "separator",
            "positive electrode",
            "current collector"
        ]:
            raise pybamm.DomainError(
                """Secondary broadcast from particle size domain must be to
                electrode or separator or current collector domains"""
            )
        elif child.domain[0] in [
            "negative electrode",
            "separator",
            "positive electrode",
        ] and broadcast_domain != ["current collector"]:
            raise pybamm.DomainError(
                """Secondary broadcast from electrode or separator must be to
                current collector domains"""
            )
        elif child.domain == ["current collector"]:
            raise pybamm.DomainError(
                "Cannot do secondary broadcast from current collector domain"
            )
        # Domain stays the same as child domain and broadcast domain is secondary
        # domain
        domain = child.domain
        auxiliary_domains = {"secondary": broadcast_domain}
        # Child's secondary domain becomes tertiary domain
        if "secondary" in child.auxiliary_domains:
            auxiliary_domains["tertiary"] = child.auxiliary_domains["secondary"]
            # Child's tertiary becomes quaternary
            if "tertiary" in child.auxiliary_domains:
                auxiliary_domains["quaternary"] = child.auxiliary_domains["tertiary"]

        return domain, auxiliary_domains

    def _unary_new_copy(self, child):
        """See :meth:`pybamm.UnaryOperator._unary_new_copy()`."""
        return SecondaryBroadcast(child, self.broadcast_domain)

    def _evaluate_for_shape(self):
        """
        Returns a vector of NaNs to represent the shape of a Broadcast.
        See :meth:`pybamm.Symbol.evaluate_for_shape_using_domain()`
        """
        child_eval = self.children[0].evaluate_for_shape()
        vec = pybamm.evaluate_for_shape_using_domain(self.domain)
        return np.outer(vec, child_eval).reshape(-1, 1)

    def reduce_one_dimension(self):
        """Reduce the broadcast by one dimension."""
        raise NotImplementedError


class SecondaryBroadcastToEdges(SecondaryBroadcast):
    """A secondary broadcast onto the edges of a domain."""

    def __init__(self, child, broadcast_domain, name=None):
        name = name or "broadcast to edges"
        super().__init__(child, broadcast_domain, name)
        self.broadcast_type = "secondary to edges"

    def _evaluates_on_edges(self, dimension):
        return True


class TertiaryBroadcast(Broadcast):
    """
    A node in the expression tree representing a tertiary broadcasting operator.
    Broadcasts in a `tertiary` dimension only. That is, makes explicit copies of the
    symbol in the domain specified by `broadcast_domain`. This is used, e.g., for
    broadcasting particle concentrations c_s(r,R) in the MPM, which have a `primary`
    and `secondary` domain, to the electrode x, which is added as a `tertiary`
    domain. Note: the symbol for broadcast must already have a non-empty `secondary`
    domain.

    Parameters
    ----------
    child : :class:`Symbol`
        child node
    broadcast_domain : iterable of str
        The domain for broadcast. This will become the tertiary domain of the symbol.
        The `tertiary` domain of the child, if present, is shifted by one to the
        `quaternary` domain of the symbol.
    name : str
        name of the node

    **Extends:** :class:`Broadcast`
    """

    def __init__(self, child, broadcast_domain, name=None):
        super().__init__(
            child, broadcast_domain, broadcast_type="tertiary to nodes", name=name
        )

    def check_and_set_domains(
        self, child, broadcast_type, broadcast_domain, broadcast_auxiliary_domains
    ):
        """See :meth:`Broadcast.check_and_set_domains`"""
        if child.auxiliary_domains == {}:
            raise TypeError(
                """Cannot take TertiaryBroadcast of an object without a secondary
                domain. Use SecondaryBroadcast instead."""
            )
        # Can only do tertiary broadcast to a "higher dimension" than the
        # secondary domain of child
        if child.auxiliary_domains["secondary"][0] in [
            "negative particle size",
            "positive particle size",
        ] and broadcast_domain[0] not in [
            "negative electrode",
            "separator",
            "positive electrode",
            "current collector"
        ]:
            raise pybamm.DomainError(
                """Tertiary broadcast from a symbol with particle size secondary
                domain must be to electrode, separator or current collector"""
            )
        if child.auxiliary_domains["secondary"][0] in [
            "negative electrode",
            "separator",
            "positive electrode",
        ] and broadcast_domain != ["current collector"]:
            raise pybamm.DomainError(
                """Tertiary broadcast from a symbol with an electrode or
                separator secondary domain must be to current collector"""
            )
        if child.auxiliary_domains["secondary"] == ["current collector"]:
            raise pybamm.DomainError(
                """Cannot do tertiary broadcast for symbol with a current collector
                secondary domain"""
            )
        # Primary and secondary domains stay the same as child's,
        # and broadcast domain is tertiary
        domain = child.domain
        auxiliary_domains = {"secondary": child.auxiliary_domains["secondary"]}
        auxiliary_domains["tertiary"] = broadcast_domain
        # Child's tertiary domain becomes quaternary
        if "tertiary" in child.auxiliary_domains:
            auxiliary_domains["quaternary"] = child.auxiliary_domains["tertiary"]

        return domain, auxiliary_domains

    def _unary_new_copy(self, child):
        """See :meth:`pybamm.UnaryOperator._unary_new_copy()`."""
        return self.__class__(child, self.broadcast_domain)

    def _evaluate_for_shape(self):
        """
        Returns a vector of NaNs to represent the shape of a Broadcast.
        See :meth:`pybamm.Symbol.evaluate_for_shape_using_domain()`
        """
        child_eval = self.children[0].evaluate_for_shape()
        vec = pybamm.evaluate_for_shape_using_domain(self.domain)
        return np.outer(vec, child_eval).reshape(-1, 1)

    def reduce_one_dimension(self):
        """Reduce the broadcast by one dimension."""
        raise NotImplementedError


class TertiaryBroadcastToEdges(TertiaryBroadcast):
    """A tertiary broadcast onto the edges of a domain."""

    def __init__(self, child, broadcast_domain, name=None):
        name = name or "broadcast to edges"
        super().__init__(child, broadcast_domain, name)
        self.broadcast_type = "tertiary to edges"

    def _evaluates_on_edges(self, dimension):
        return True


class FullBroadcast(Broadcast):
    """A class for full broadcasts."""

    def __init__(self, child, broadcast_domain, auxiliary_domains, name=None):
        if isinstance(auxiliary_domains, str):
            auxiliary_domains = {"secondary": auxiliary_domains}
        super().__init__(
            child,
            broadcast_domain,
            broadcast_auxiliary_domains=auxiliary_domains,
            broadcast_type="full to nodes",
            name=name,
        )

    def check_and_set_domains(
        self, child, broadcast_type, broadcast_domain, broadcast_auxiliary_domains
    ):
        """See :meth:`Broadcast.check_and_set_domains`"""

        # Variables on the current collector can only be broadcast to 'primary'
        if child.domain == ["current collector"]:
            raise pybamm.DomainError(
                "Cannot do full broadcast from current collector domain"
            )
        domain = broadcast_domain
        auxiliary_domains = broadcast_auxiliary_domains or {}

        return domain, auxiliary_domains

    def _unary_new_copy(self, child):
        """See :meth:`pybamm.UnaryOperator._unary_new_copy()`."""
        return self.__class__(child, self.broadcast_domain, self.auxiliary_domains)

    def _evaluate_for_shape(self):
        """
        Returns a vector of NaNs to represent the shape of a Broadcast.
        See :meth:`pybamm.Symbol.evaluate_for_shape_using_domain()`
        """
        child_eval = self.children[0].evaluate_for_shape()
        vec = pybamm.evaluate_for_shape_using_domain(
            self.domain, self.auxiliary_domains
        )

        return child_eval * vec

    def reduce_one_dimension(self):
        """Reduce the broadcast by one dimension."""
        if self.auxiliary_domains == {}:
            return self.orphans[0]
        elif "tertiary" not in self.auxiliary_domains:
            return PrimaryBroadcast(
                self.orphans[0], self.auxiliary_domains["secondary"]
            )
        elif "tertiary" in self.auxiliary_domains:
            return FullBroadcast(
                self.orphans[0],
                self.auxiliary_domains["secondary"],
                {"secondary": self.auxiliary_domains["tertiary"]},
            )


class FullBroadcastToEdges(FullBroadcast):
    """
    A full broadcast onto the edges of a domain (edges of primary dimension, nodes of
    other dimensions)
    """

    def __init__(self, child, broadcast_domain, auxiliary_domains, name=None):
        name = name or "broadcast to edges"
        super().__init__(child, broadcast_domain, auxiliary_domains, name)
        self.broadcast_type = "full to edges"

    def _evaluates_on_edges(self, dimension):
        return True

    def reduce_one_dimension(self):
        """Reduce the broadcast by one dimension."""
        if self.auxiliary_domains == {}:
            return self.orphans[0]
        elif "tertiary" not in self.auxiliary_domains:
            return PrimaryBroadcastToEdges(
                self.orphans[0], self.auxiliary_domains["secondary"]
            )
        elif "tertiary" in self.auxiliary_domains:
            return FullBroadcastToEdges(
                self.orphans[0],
                self.auxiliary_domains["secondary"],
                {"secondary": self.auxiliary_domains["tertiary"]},
            )


def full_like(symbols, fill_value):
    """
    Returns an array with the same shape, domain and auxiliary domains as the sum of the
    input symbols, with a constant value given by `fill_value`.

    Parameters
    ----------
    symbols : :class:`Symbol`
        Symbols whose shape to copy
    fill_value : number
        Value to assign
    """
    # Make a symbol that combines all the children, to get the right domain
    # that takes all the child symbols into account
    sum_symbol = symbols[0]
    for sym in symbols[1:]:
        sum_symbol += sym

    # Just return scalar if symbol shape is scalar
    if sum_symbol.evaluates_to_number():
        return pybamm.Scalar(fill_value)
    try:
        shape = sum_symbol.shape
        # use vector or matrix
        if shape[1] == 1:
            array_type = pybamm.Vector
        else:
            array_type = pybamm.Matrix
        # return dense array, except for a matrix of zeros
        if shape[1] != 1 and fill_value == 0:
            entries = csr_matrix(shape)
        else:
            entries = fill_value * np.ones(shape)

        return array_type(
            entries,
            domain=sum_symbol.domain,
            auxiliary_domains=sum_symbol.auxiliary_domains,
        )

    except NotImplementedError:
        if sum_symbol.evaluates_on_edges("primary"):
            return FullBroadcastToEdges(
                fill_value, sum_symbol.domain, sum_symbol.auxiliary_domains
            )
        else:
            return FullBroadcast(
                fill_value, sum_symbol.domain, sum_symbol.auxiliary_domains
            )


def zeros_like(*symbols):
    """
    Returns an array with the same shape, domain and auxiliary domains as the sum of the
    input symbols, with each entry equal to zero.

    Parameters
    ----------
    symbols : :class:`Symbol`
        Symbols whose shape to copy
    """
    return full_like(symbols, 0)


def ones_like(*symbols):
    """
    Returns an array with the same shape, domain and auxiliary domains as the sum of the
    input symbols, with each entry equal to one.

    Parameters
    ----------
    symbols : :class:`Symbol`
        Symbols whose shape to copy
    """
    return full_like(symbols, 1)



#
# Class for reaction limited SEI growth
#
import pybamm
from .base_sei import BaseModel


class EcReactionLimited(BaseModel):
    """
    Class for reaction limited SEI growth. This model assumes the "inner"
    SEI layer is of zero thickness and only models the "outer" SEI layer.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    reaction_loc : str
        Where the reaction happens: "x-average" (SPM, SPMe, etc),
        "full electrode" (full DFN), or "interface" (half-cell DFN)
    options : dict, optional
        A dictionary of options to be passed to the model.

    **Extends:** :class:`pybamm.sei.BaseModel`
    """

    def __init__(self, param, reaction_loc, options=None):
        super().__init__(param, options=options)
        self.reaction_loc = reaction_loc

    def get_fundamental_variables(self):

        if self.reaction_loc == "x-average":
            L_inner = pybamm.FullBroadcast(0, "negative electrode", "current collector")
            L_outer_av = pybamm.standard_variables.L_outer_av
            L_outer = pybamm.PrimaryBroadcast(L_outer_av, "negative electrode")
        elif self.reaction_loc == "full electrode":
            L_inner = pybamm.FullBroadcast(0, "negative electrode", "current collector")
            L_outer = pybamm.standard_variables.L_outer
        elif self.reaction_loc == "interface":
            L_inner = pybamm.PrimaryBroadcast(0, "current collector")
            L_outer = pybamm.standard_variables.L_outer_interface

        variables = self._get_standard_thickness_variables(L_inner, L_outer)
        variables.update(self._get_standard_concentration_variables(variables))

        return variables

    def get_coupled_variables(self, variables):
        # delta_phi = phi_s - phi_e
        if self.reaction_loc == "interface":
            delta_phi = variables[
                "Lithium metal interface surface potential difference"
            ]
        else:
            delta_phi = variables["Negative electrode surface potential difference"]

        L_sei = variables["Outer SEI thickness"]

        # Look for current that contributes to the -IR drop
        # If we can't find the interfacial current density from the main reaction, j,
        # it's ok to fall back on the total interfacial current density, j_tot
        # This should only happen when the interface submodel is "InverseButlerVolmer"
        # in which case j = j_tot (uniform) anyway
        if "Negative electrode interfacial current density" in variables:
            j = variables["Negative electrode interfacial current density"]
        elif self.reaction_loc == "interface":
            j = variables["Lithium metal total interfacial current density"]
        else:
            j = variables[
                "X-averaged "
                + self.domain.lower()
                + " electrode total interfacial current density"
            ]

        C_sei_ec = self.param.C_sei_ec
        R_sei = self.param.R_sei
        C_ec = self.param.C_ec

        # we have a linear system for j_sei and c_ec
        #  c_ec = 1 + j_sei * L_sei * C_ec
        #  j_sei = - C_sei_ec * c_ec * exp()
        # so
        #  j_sei = - C_sei_ec * exp() - j_sei * L_sei * C_ec * C_sei_ec * exp()
        # so
        #  j_sei = -C_sei_ec * exp() / (1 + L_sei * C_ec * C_sei_ec * exp())
        #  c_ec = 1 / (1 + L_sei * C_ec * C_sei_ec * exp())
        # need to revise for thermal case
        C_sei_exp = C_sei_ec * pybamm.exp(-0.5 * (delta_phi - j * L_sei * R_sei))
        j_sei = -C_sei_exp / (1 + L_sei * C_ec * C_sei_exp)
        c_ec = 1 / (1 + L_sei * C_ec * C_sei_exp)

        if self.reaction_loc == "interface":
            j_inner = pybamm.PrimaryBroadcast(0, "current collector")
        else:
            j_inner = pybamm.FullBroadcast(0, "negative electrode", "current collector")
        j_outer = j_sei

        variables.update(self._get_standard_reaction_variables(j_inner, j_outer))

        # Get variables related to the concentration
        c_ec_av = pybamm.x_average(c_ec)
        c_ec_scale = self.param.c_ec_0_dim

        variables.update(
            {
                "EC surface concentration": c_ec,
                "EC surface concentration [mol.m-3]": c_ec * c_ec_scale,
                "X-averaged EC surface concentration": c_ec_av,
                "X-averaged EC surface concentration [mol.m-3]": c_ec_av * c_ec_scale,
            }
        )

        # Update whole cell variables, which also updates the "sum of" variables
        variables.update(super().get_coupled_variables(variables))

        return variables

    def set_rhs(self, variables):
        if self.reaction_loc == "x-average":
            L_sei = variables["X-averaged outer SEI thickness"]
            j_sei = variables["X-averaged outer SEI interfacial current density"]
        else:
            L_sei = variables["Outer SEI thickness"]
            j_sei = variables["Outer SEI interfacial current density"]

        Gamma_SEI = self.param.Gamma_SEI

        self.rhs = {L_sei: -Gamma_SEI * j_sei / 2}

    def set_initial_conditions(self, variables):
        if self.reaction_loc == "x-average":
            L_sei = variables["X-averaged outer SEI thickness"]
        else:
            L_sei = variables["Outer SEI thickness"]
        L_sei_0 = pybamm.Scalar(1)

        self.initial_conditions = {L_sei: L_sei_0}



def copper_thermal_conductivity_CRC(T):
    """
    Copper thermal conductivity as a function of the temperature from [1].

    References
    ----------
    .. [1] William M. Haynes (Ed.). "CRC handbook of chemistry and physics". CRC Press
    (2014).

    Parameters
    ----------
    T: :class:`pybamm.Symbol`
       Dimensional temperature

    Returns
    -------
    :class:`pybamm.Symbol`
       Thermal conductivity
    """

    lambda_th = -5.409e-7 * T ** 3 + 7.054e-4 * T ** 2 - 0.3727 * T + 463.6

    return lambda_th



from pybamm import exp, constants


def nmc_LGM50_diffusivity_ORegan2021(sto, T):
    """
    NMC diffusivity as a function of stoichiometry, in this case the
    diffusivity is taken to be a constant. The value is taken from [1].

    References
    ----------
    .. [1] Kieran O’Regan, Ferran Brosa Planella, W. Dhammika Widanage, and Emma
    Kendrick. "Thermal-electrochemical parametrisation of a lithium-ion battery:
    mapping Li concentration and temperature dependencies." Journal of the
    Electrochemical Society, submitted (2021).

    Parameters
    ----------
    sto: :class:`pybamm.Symbol`
       Electrode stochiometry
    T: :class:`pybamm.Symbol`
       Dimensional temperature

    Returns
    -------
    :class:`pybamm.Symbol`
       Solid diffusivity
    """

    a1 = -0.9231
    a2 = -0.4066
    a3 = -0.993
    b1 = 0.3216
    b2 = 0.4532
    b3 = 0.8098
    c0 = -13.96
    c1 = 0.002534
    c2 = 0.003926
    c3 = 0.09924
    d = 1449

    D_ref = (
        10
        ** (
            c0
            + a1 * exp(-((sto - b1) ** 2) / c1)
            + a2 * exp(-((sto - b2) ** 2) / c2)
            + a3 * exp(-((sto - b3) ** 2) / c3)
        )
        * 2.7  # correcting factor (see O'Regan et al 2021)
    )

    E_D_s = d * constants.R
    arrhenius = exp(E_D_s / constants.R * (1 / 298.15 - 1 / T))

    return D_ref * arrhenius



from pybamm import exp, constants, Parameter


def graphite_electrolyte_exchange_current_density_Kim2011(c_e, c_s_surf, T):
    """
    Exchange-current density for Butler-Volmer reactions between graphite and LiPF6 in
    EC:DMC
    [1].

    References
    ----------
    .. [1] Kim, G. H., Smith, K., Lee, K. J., Santhanagopalan, S., & Pesaran, A.
    (2011). Multi-domain modeling of lithium-ion batteries encompassing
    multi-physics in varied length scales. Journal of The Electrochemical
    Society, 158(8), A955-A969.

    Parameters
    ----------
    c_e : :class:`pybamm.Symbol`
        Electrolyte concentration [mol.m-3]
    c_s_surf : :class:`pybamm.Symbol`
        Particle concentration [mol.m-3]
    T : :class:`pybamm.Symbol`
        Temperature [K]

    Returns
    -------
    :class:`pybamm.Symbol`
        Exchange-current density [A.m-2]
    """

    i0_ref = 36  # reference exchange current density at 100% SOC
    sto = 0.36  # stochiometry at 100% SOC
    c_s_n_max = Parameter("Maximum concentration in negative electrode [mol.m-3]")
    c_s_n_ref = sto * c_s_n_max  # reference electrode concentration
    c_e_ref = Parameter("Typical electrolyte concentration [mol.m-3]")
    alpha = 0.5  # charge transfer coefficient

    m_ref = i0_ref / (
        c_e_ref ** alpha * (c_s_n_max - c_s_n_ref) ** alpha * c_s_n_ref ** alpha
    )

    E_r = 3e4
    arrhenius = exp(E_r / constants.R * (1 / 298.15 - 1 / T))

    return (
        m_ref
        * arrhenius
        * c_e ** alpha
        * c_s_surf ** alpha
        * (c_s_n_max - c_s_surf) ** alpha
    )



#
# Base model for the external circuit
#
import pybamm


class BaseModel(pybamm.BaseSubModel):
    """Model to represent the behaviour of the external circuit."""

    def __init__(self, param):
        super().__init__(param)

    def get_fundamental_variables(self):
        Q = pybamm.standard_variables.Q
        variables = {"Discharge capacity [A.h]": Q}
        return variables

    def set_initial_conditions(self, variables):
        Q = variables["Discharge capacity [A.h]"]
        self.initial_conditions[Q] = pybamm.Scalar(0)

    def set_rhs(self, variables):
        # ODE for discharge capacity
        Q = variables["Discharge capacity [A.h]"]
        I = variables["Current [A]"]
        self.rhs[Q] = I * self.param.timescale / 3600


class LeadingOrderBaseModel(BaseModel):
    """Model to represent the behaviour of the external circuit, at leading order."""

    def __init__(self, param):
        super().__init__(param)

    def get_fundamental_variables(self):
        Q = pybamm.Variable("Leading-order discharge capacity [A.h]")
        variables = {"Discharge capacity [A.h]": Q}
        return variables



#
# Base class for porosity
#
import pybamm


class BaseModel(pybamm.BaseSubModel):
    """Base class for porosity

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel


    **Extends:** :class:`pybamm.BaseSubModel`
    """

    def __init__(self, param, options):
        super().__init__(param, options=options)

    def _get_standard_porosity_variables(
        self, eps_n, eps_s, eps_p, set_leading_order=False
    ):
        if not self.half_cell:
            eps_n_av = pybamm.x_average(eps_n)

        eps_s_av = pybamm.x_average(eps_s)
        eps_p_av = pybamm.x_average(eps_p)
        eps = pybamm.concatenation(eps_n, eps_s, eps_p)

        variables = {
            "Porosity": eps,
            "Separator porosity": eps_s,
            "Positive electrode porosity": eps_p,
            "X-averaged separator porosity": eps_s_av,
            "X-averaged positive electrode porosity": eps_p_av,
        }

        if not self.half_cell:
            variables.update(
                {
                    "Negative electrode porosity": eps_n,
                    "X-averaged negative electrode porosity": eps_n_av,
                }
            )

        if set_leading_order is True:
            leading_order_variables = {
                "Leading-order " + name.lower(): var for name, var in variables.items()
            }
            variables.update(leading_order_variables)

        return variables

    def _get_standard_porosity_change_variables(
        self, deps_n_dt, deps_s_dt, deps_p_dt, set_leading_order=False
    ):
        if deps_p_dt.domain == ["current collector"]:
            if not self.half_cell:
                deps_n_dt_av = deps_n_dt
                deps_n_dt = pybamm.PrimaryBroadcast(deps_n_dt_av, "negative electrode")
            deps_s_dt_av = deps_s_dt
            deps_p_dt_av = deps_p_dt
            deps_s_dt = pybamm.PrimaryBroadcast(deps_s_dt_av, "separator")
            deps_p_dt = pybamm.PrimaryBroadcast(deps_p_dt_av, "positive electrode")
        else:
            if not self.half_cell:
                deps_n_dt_av = pybamm.x_average(deps_n_dt)
            deps_s_dt_av = pybamm.x_average(deps_s_dt)
            deps_p_dt_av = pybamm.x_average(deps_p_dt)
        deps_dt = pybamm.concatenation(deps_n_dt, deps_s_dt, deps_p_dt)

        variables = {
            "Porosity change": deps_dt,
            "Separator porosity change": deps_s_dt,
            "Positive electrode porosity change": deps_p_dt,
            "X-averaged separator porosity change": deps_s_dt_av,
            "X-averaged positive electrode porosity change": deps_p_dt_av,
        }

        if not self.half_cell:
            variables.update(
                {
                    "Negative electrode porosity change": deps_n_dt,
                    "X-averaged negative electrode porosity change": deps_n_dt_av,
                }
            )

        if set_leading_order is True:
            variables.update(
                {
                    "Leading-order x-averaged separator porosity change": deps_s_dt_av,
                    "Leading-order x-averaged "
                    + "positive electrode porosity change": deps_p_dt_av,
                }
            )
            if not self.half_cell:
                variables.update(
                    {
                        "Leading-order x-averaged "
                        + "negative electrode porosity change": deps_n_dt_av,
                    }
                )

        return variables

    def set_events(self, variables):
        if not self.half_cell:
            eps_n = variables["Negative electrode porosity"]
            self.events.append(
                pybamm.Event(
                    "Zero negative electrode porosity cut-off",
                    pybamm.min(eps_n),
                    pybamm.EventType.TERMINATION,
                )
            )
            self.events.append(
                pybamm.Event(
                    "Max negative electrode porosity cut-off",
                    pybamm.max(eps_n) - 1,
                    pybamm.EventType.TERMINATION,
                )
            )

        eps_p = variables["Positive electrode porosity"]
        self.events.append(
            pybamm.Event(
                "Zero positive electrode porosity cut-off",
                pybamm.min(eps_p),
                pybamm.EventType.TERMINATION,
            )
        )
        self.events.append(
            pybamm.Event(
                "Max positive electrode porosity cut-off",
                pybamm.max(eps_p) - 1,
                pybamm.EventType.TERMINATION,
            )
        )



def electrolyte_diffusivity_Valoen2005(c_e, T):
    """
    Diffusivity of LiPF6 in EC:DMC as a function of ion concentration, from [1] (eqn 14)

    References
    ----------
    .. [1] Valøen, Lars Ole, and Jan N. Reimers. "Transport properties of LiPF6-based
    Li-ion battery electrolytes." Journal of The Electrochemical Society 152.5 (2005):
    A882-A891.

    Parameters
    ----------
    c_e : :class:`pybamm.Symbol`
        Dimensional electrolyte concentration [mol.m-3]
    T : :class:`pybamm.Symbol`
        Dimensional temperature [K]

    Returns
    -------
    :class:`pybamm.Symbol`
        Dimensional electrolyte diffusivity [m2.s-1]
    """
    # mol/m3 to molar
    c_e = c_e / 1000

    T_g = 229 + 5 * c_e
    D_0 = -4.43 - 54 / (T - T_g)
    D_1 = -0.22

    # cm2/s to m2/s
    # note, in the Valoen paper, ln means log10, so its inverse is 10^x
    return (10 ** (D_0 + D_1 * c_e)) * 1e-4



#
# Class for solvent-diffusion limited SEI growth
#
import pybamm
from .base_sei import BaseModel


class SolventDiffusionLimited(BaseModel):
    """
    Class for solvent-diffusion limited SEI growth.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    reaction_loc : str
        Where the reaction happens: "x-average" (SPM, SPMe, etc),
        "full electrode" (full DFN), or "interface" (half-cell DFN)
    options : dict, optional
        A dictionary of options to be passed to the model.

    **Extends:** :class:`pybamm.sei.BaseModel`
    """

    def __init__(self, param, reaction_loc, options=None):
        super().__init__(param, options=options)
        self.reaction_loc = reaction_loc

    def get_fundamental_variables(self):
        if self.reaction_loc == "x-average":
            L_inner_av = pybamm.standard_variables.L_inner_av
            L_outer_av = pybamm.standard_variables.L_outer_av
            L_inner = pybamm.PrimaryBroadcast(L_inner_av, "negative electrode")
            L_outer = pybamm.PrimaryBroadcast(L_outer_av, "negative electrode")
        elif self.reaction_loc == "full electrode":
            L_inner = pybamm.standard_variables.L_inner
            L_outer = pybamm.standard_variables.L_outer
        elif self.reaction_loc == "interface":
            L_inner = pybamm.standard_variables.L_inner_interface
            L_outer = pybamm.standard_variables.L_outer_interface

        variables = self._get_standard_thickness_variables(L_inner, L_outer)
        variables.update(self._get_standard_concentration_variables(variables))

        return variables

    def get_coupled_variables(self, variables):
        L_sei_outer = variables["Outer SEI thickness"]

        C_sei = self.param.C_sei_solvent

        j_sei = -1 / (C_sei * L_sei_outer)

        alpha = 0.5
        j_inner = alpha * j_sei
        j_outer = (1 - alpha) * j_sei

        variables.update(self._get_standard_reaction_variables(j_inner, j_outer))

        # Update whole cell variables, which also updates the "sum of" variables
        variables.update(super().get_coupled_variables(variables))

        return variables

    def set_rhs(self, variables):
        if self.reaction_loc == "x-average":
            L_inner = variables["X-averaged inner SEI thickness"]
            L_outer = variables["X-averaged outer SEI thickness"]
            j_inner = variables["X-averaged inner SEI interfacial current density"]
            j_outer = variables["X-averaged outer SEI interfacial current density"]
        else:
            L_inner = variables["Inner SEI thickness"]
            L_outer = variables["Outer SEI thickness"]
            j_inner = variables["Inner SEI interfacial current density"]
            j_outer = variables["Outer SEI interfacial current density"]
        v_bar = self.param.v_bar

        Gamma_SEI = self.param.Gamma_SEI

        self.rhs = {
            L_inner: -Gamma_SEI * j_inner,
            L_outer: -v_bar * Gamma_SEI * j_outer,
        }

    def set_initial_conditions(self, variables):
        if self.reaction_loc == "x-average":
            L_inner = variables["X-averaged inner SEI thickness"]
            L_outer = variables["X-averaged outer SEI thickness"]
        else:
            L_inner = variables["Inner SEI thickness"]
            L_outer = variables["Outer SEI thickness"]
        L_inner_0 = self.param.L_inner_0
        L_outer_0 = self.param.L_outer_0

        self.initial_conditions = {L_inner: L_inner_0, L_outer: L_outer_0}



def nmc_LGM50_thermal_conductivity_ORegan2021(T):
    """
    Wet positive electrode thermal conductivity as a function of the temperature from
    [1].

    References
    ----------
    .. [1] Kieran O’Regan, Ferran Brosa Planella, W. Dhammika Widanage, and Emma
    Kendrick. "Thermal-electrochemical parametrisation of a lithium-ion battery:
    mapping Li concentration and temperature dependencies." Journal of the
    Electrochemical Society, submitted (2021).

    Parameters
    ----------
    T: :class:`pybamm.Symbol`
       Dimensional temperature

    Returns
    -------
    :class:`pybamm.Symbol`
       Thermal conductivity
    """

    lambda_wet = 2.063e-5 * T ** 2 - 0.01127 * T + 2.331

    return lambda_wet



#
# Class for reaction driven porosity changes as a multiple of SEI/plating thicknesses
#
from .base_porosity import BaseModel


class ReactionDriven(BaseModel):
    """Reaction-driven porosity changes as a multiple of SEI/plating thicknesses

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    options : dict
        Options dictionary passed from the full model
    x_average : bool
        Whether to use x-averaged variables (SPM, SPMe, etc) or full variables (DFN)

    **Extends:** :class:`pybamm.porosity.BaseModel`
    """

    def __init__(self, param, options, x_average):
        super().__init__(param, options)
        self.x_average = x_average

    def get_coupled_variables(self, variables):
        L_sei_n = variables["Total SEI thickness [m]"]
        L_sei_0 = self.param.L_inner_0_dim + self.param.L_outer_0_dim
        L_pl_n = variables["Lithium plating thickness [m]"]

        L_tot = (L_sei_n - L_sei_0) + L_pl_n

        a_n = variables["Negative electrode surface area to volume ratio [m-1]"]

        # This assumes a thin film so curvature effects are neglected. They could be
        # included (i.e. for a sphere it is a_n * (L_tot + L_tot ** 2 / R_n + L_tot **
        # 3 / (3 * R_n ** 2))) but it is not clear if it is relevant or not.
        delta_eps_n = -a_n * L_tot

        eps_n = self.param.epsilon_n_init + delta_eps_n
        eps_s = self.param.epsilon_s_init
        # no SEI or plating in the positive electrode
        eps_p = self.param.epsilon_p_init

        variables = self._get_standard_porosity_variables(eps_n, eps_s, eps_p)

        return variables



from pybamm import exp, constants


def electrolyte_conductivity_PeymanMPM(c_e, T):
    """
    Conductivity of LiPF6 in EC:DMC as a function of ion concentration. The original
    data is from [1]. The fit is from Dualfoil [2].

    References
    ----------
    .. [1] C Capiglia et al. 7Li and 19F diffusion coefficients and thermal
    properties of non-aqueous electrolyte solutions for rechargeable lithium batteries.
    Journal of power sources 81 (1999): 859-862.
    .. [2] http://www.cchem.berkeley.edu/jsngrp/fortran.html
    Parameters
    ----------
    c_e: :class:`pybamm.Symbol`
        Dimensional electrolyte concentration
    T: :class:`pybamm.Symbol`
        Dimensional temperature


    Returns
    -------
    :class:`pybamm.Symbol`
        Electrolyte conductivity
    """

    sigma_e = 1.3
    E_k_e = 34700
    arrhenius = exp(E_k_e / constants.R * (1 / 298.15 - 1 / T))

    return sigma_e * arrhenius



from pybamm import exp, constants, Parameter


def graphite_LGM50_electrolyte_exchange_current_density_ORegan2021(c_e, c_s_surf, T):
    """
    Exchange-current density for Butler-Volmer reactions between graphite and LiPF6 in
    EC:DMC.

    References
    ----------
    .. [1] Kieran O’Regan, Ferran Brosa Planella, W. Dhammika Widanage, and Emma
    Kendrick. "Thermal-electrochemical parametrisation of a lithium-ion battery:
    mapping Li concentration and temperature dependencies." Journal of the
    Electrochemical Society, submitted (2021).

    Parameters
    ----------
    c_e : :class:`pybamm.Symbol`
        Electrolyte concentration [mol.m-3]
    c_s_surf : :class:`pybamm.Symbol`
        Particle concentration [mol.m-3]
    T : :class:`pybamm.Symbol`
        Temperature [K]

    Returns
    -------
    :class:`pybamm.Symbol`
        Exchange-current density [A.m-2]
    """

    i_ref = 2.668  # (A/m2)
    alpha = 0.792
    E_r = 4e4
    arrhenius = exp(E_r / constants.R * (1 / 298.15 - 1 / T))

    c_n_max = Parameter("Maximum concentration in negative electrode [mol.m-3]")
    c_e_ref = Parameter("Typical electrolyte concentration [mol.m-3]")

    return (
        i_ref
        * arrhenius
        * (c_e / c_e_ref) ** (1 - alpha)
        * (c_s_surf / c_n_max) ** alpha
        * (1 - c_s_surf / c_n_max) ** (1 - alpha)
    )



#
# Processed Variable class
#
import casadi
import numbers
import numpy as np


class ProcessedSymbolicVariable(object):
    """
    An object that can be evaluated at arbitrary (scalars or vectors) t and x, and
    returns the (interpolated) value of the base variable at that t and x.

    Parameters
    ----------
    base_variable : :class:`pybamm.Symbol`
        A base variable with a method `evaluate(t,y)` that returns the value of that
        variable. Note that this can be any kind of node in the expression tree, not
        just a :class:`pybamm.Variable`.
        When evaluated, returns an array of size (m,n)
    solution : :class:`pybamm.Solution`
        The solution object to be used to create the processed variables
    """

    def __init__(self, base_variable, solution):
        # Convert variable to casadi
        t_MX = casadi.MX.sym("t")
        y_MX = casadi.MX.sym("y", solution.all_ys[0].shape[0])
        # Make all inputs symbolic first for converting to casadi
        all_inputs_as_MX_dict = {}
        symbolic_inputs_dict = {}
        for key, value in solution.all_inputs[0].items():
            if not isinstance(value, casadi.MX):
                all_inputs_as_MX_dict[key] = casadi.MX.sym("input")
            else:
                all_inputs_as_MX_dict[key] = value
                # Only add symbolic inputs to the "symbolic_inputs" dict
                symbolic_inputs_dict[key] = value

        all_inputs_as_MX = casadi.vertcat(*[p for p in all_inputs_as_MX_dict.values()])
        # The symbolic_inputs dictionary will be used for sensitivity
        symbolic_inputs = casadi.vertcat(*[p for p in symbolic_inputs_dict.values()])
        var = base_variable.to_casadi(t_MX, y_MX, inputs=all_inputs_as_MX_dict)

        self.base_variable = casadi.Function(
            "variable", [t_MX, y_MX, all_inputs_as_MX], [var]
        )
        # Store some attributes
        self.t_pts = solution.t
        self.all_ts = solution.all_ts
        self.all_ys = solution.all_ys
        self.mesh = base_variable.mesh
        self.all_inputs_casadi = solution.all_inputs_casadi

        self.symbolic_inputs_dict = symbolic_inputs_dict
        self.symbolic_inputs_total_shape = symbolic_inputs.shape[0]
        self.domain = base_variable.domain

        self.base_eval = self.base_variable(
            self.all_ts[0][0], self.all_ys[0][:, 0], self.all_inputs_casadi[0]
        )

        if (
            isinstance(self.base_eval, numbers.Number)
            or len(self.base_eval.shape) == 0
            or self.base_eval.shape[0] == 1
        ):
            self.initialise_0D()
        else:
            n = self.mesh.npts
            base_shape = self.base_eval.shape[0]
            # Try shape that could make the variable a 1D variable
            if base_shape == n:
                self.initialise_1D()
            else:
                # Raise error for 2D variable
                raise NotImplementedError(
                    "Shape not recognized for {} ".format(base_variable)
                    + "(note processing of 2D and 3D variables is not yet "
                    + "implemented)"
                )

        # Make entries a function and compute jacobian
        entries_MX = self.entries
        self.casadi_entries_fn = casadi.Function(
            "variable", [symbolic_inputs], [entries_MX]
        )

        # Don't compute jacobian if the entries are a DM (not symbolic)
        if isinstance(entries_MX, casadi.DM):
            self.casadi_sens_fn = None
        # Do compute jacobian if the entries are symbolic (functions of input)
        else:
            sens_MX = casadi.jacobian(entries_MX, symbolic_inputs)
            self.casadi_sens_fn = casadi.Function(
                "variable", [symbolic_inputs], [sens_MX]
            )

    def initialise_0D(self):
        """Create a 0D variable"""
        # Evaluate the base_variable index-by-index
        idx = 0
        for ts, ys, inputs in zip(self.all_ts, self.all_ys, self.all_inputs_casadi):
            for inner_idx, t in enumerate(ts):
                t = ts[inner_idx]
                y = ys[:, inner_idx]
                next_entries = self.base_variable(t, y, inputs)
                if idx == 0:
                    entries = next_entries
                else:
                    entries = casadi.horzcat(entries, next_entries)
                idx += 1

        self.entries = entries
        self.dimensions = 0

    def initialise_1D(self):
        """Create a 1D variable"""
        len_space = self.base_eval.shape[0]
        entries = np.empty((len_space, len(self.t_pts)))

        # Evaluate the base_variable index-by-index
        idx = 0
        for ts, ys, inputs in zip(self.all_ts, self.all_ys, self.all_inputs_casadi):
            for inner_idx, t in enumerate(ts):
                t = ts[inner_idx]
                y = ys[:, inner_idx]
                next_entries = self.base_variable(t, y, inputs)
                if idx == 0:
                    entries = next_entries
                else:
                    entries = casadi.vertcat(entries, next_entries)
                idx += 1

        self.entries = entries

    def value(self, inputs=None, check_inputs=True):
        """
        Returns the value of the variable at the specified input values

        Parameters
        ----------
        inputs : dict
            The inputs at which to evaluate the variable.

        Returns
        -------
        casadi.DM
            A casadi matrix of size (n_x * n_t, 1), where n_x is the number of spatial
            discretisation points for the variable, and n_t is the length of the time
            vector
        """
        if inputs is None:
            return self.casadi_entries_fn(casadi.DM())
        else:
            if check_inputs:
                inputs = self._check_and_transform(inputs)
            return self.casadi_entries_fn(inputs)

    def sensitivity(self, inputs=None, check_inputs=True):
        """
        Returns the sensitivity of the variable to the symbolic inputs at the specified
        input values

        Parameters
        ----------
        inputs : dict
            The inputs at which to evaluate the variable.

        Returns
        -------
        casadi.DM
            A casadi matrix of size (n_x * n_t, n_p), where n_x is the number of spatial
            discretisation points for the variable, n_t is the length of the time
            vector, and n_p is the number of input parameters
        """
        if self.casadi_sens_fn is None:
            raise ValueError(
                "Variable is not symbolic, so sensitivities are not defined"
            )
        if check_inputs:
            inputs = self._check_and_transform(inputs)
        return self.casadi_sens_fn(inputs)

    def value_and_sensitivity(self, inputs=None):
        """
        Returns the value of the variable and its sensitivity to the symbolic inputs at
        the specified input values

        Parameters
        ----------
        inputs : dict
            The inputs at which to evaluate the variable.
        """
        inputs = self._check_and_transform(inputs)
        # Pass check_inputs=False to avoid re-checking inputs
        return (
            self.value(inputs, check_inputs=False),
            self.sensitivity(inputs, check_inputs=False),
        )

    def _check_and_transform(self, inputs_dict):
        """Check dictionary has the right inputs, and convert to a vector"""
        # Convert dict to casadi vector
        if not isinstance(inputs_dict, dict):
            raise TypeError("inputs should be 'dict' but are {}".format(inputs_dict))
        # Sort input dictionary keys according to the symbolic inputs dictionary
        # For practical number of input parameters this should be extremely fast and
        # so is ok to do at each step
        try:
            inputs_dict_sorted = {
                k: inputs_dict[k] for k in self.symbolic_inputs_dict.keys()
            }
        except KeyError as e:
            raise KeyError("Inconsistent input keys. '{}' not found".format(e.args[0]))
        inputs = casadi.vertcat(*[p for p in inputs_dict_sorted.values()])
        if inputs.shape[0] != self.symbolic_inputs_total_shape:
            # Find the variable which caused the error, for a clearer error message
            for key, inp in inputs_dict_sorted.items():
                if inp.shape[0] != self.symbolic_inputs_dict[key].shape[0]:
                    raise ValueError(
                        "Wrong shape for input '{}': expected {}, actual {}".format(
                            key, self.symbolic_inputs_dict[key].shape[0], inp.shape[0]
                        )
                    )

        return inputs

    @property
    def data(self):
        """Same as entries, but different name"""
        return self.entries



from pybamm import tanh


def nmc_LGM50_ocp_Chen2020(sto):
    """
     LG M50 NMC open-circuit potential as a function of stoichiometry.  The fit is
    taken from [1].

     References
     ----------
    .. [1] Chang-Hui Chen, Ferran Brosa Planella, Kieran O’Regan, Dominika Gastol, W.
    Dhammika Widanage, and Emma Kendrick. "Development of Experimental Techniques for
    Parameterization of Multi-scale Lithium-ion Battery Models." Journal of the
    Electrochemical Society 167 (2020): 080534.

     Parameters
     ----------
     sto: :class:`pybamm.Symbol`
       Electrode stochiometry

     Returns
     -------
     :class:`pybamm.Symbol`
        Open-circuit potential
    """

    U = (
        -0.809 * sto
        + 4.4875
        - 0.0428 * tanh(18.5138 * (sto - 0.5542))
        - 17.7326 * tanh(15.789 * (sto - 0.3117))
        + 17.5842 * tanh(15.9308 * (sto - 0.312))
    )

    return U



#
# One-dimensional submeshes
#
import pybamm
from .meshes import SubMesh

import numpy as np


class SubMesh1D(SubMesh):
    """
    1D submesh class.
    Contains the position of the nodes, the number of mesh points, and
    (optionally) information about the tab locations.

    Parameters
    ----------
    edges : array_like
        An array containing the points corresponding to the edges of the submesh
    coord_sys : string
        The coordinate system of the submesh
    tabs : dict, optional
        A dictionary that contains information about the size and location of
        the tabs

    **Extends:"": :class:`pybamm.SubMesh`
    """

    def __init__(self, edges, coord_sys, tabs=None):
        self.edges = edges
        self.nodes = (self.edges[1:] + self.edges[:-1]) / 2
        self.d_edges = np.diff(self.edges)
        self.d_nodes = np.diff(self.nodes)
        self.npts = self.nodes.size
        self.coord_sys = coord_sys
        self.internal_boundaries = []

        # Add tab locations in terms of "left" and "right"
        if tabs:
            self.tabs = {}
            l_z = self.edges[-1]

            def near(x, point, tol=3e-16):
                return abs(x - point) < tol

            for tab in ["negative", "positive"]:
                tab_location = tabs[tab]["z_centre"]
                if near(tab_location, 0):
                    self.tabs[tab + " tab"] = "left"
                elif near(tab_location, l_z):
                    self.tabs[tab + " tab"] = "right"
                else:
                    raise pybamm.GeometryError(
                        """{} tab located at {}, but must be at either 0 or {}
                        (in dimensionless coordinates).""".format(
                            tab, tab_location, l_z
                        )
                    )

    def read_lims(self, lims):
        # Separate limits and tabs
        # Read and remove tabs. If "tabs" is not a key in "lims", then tabs is set to
        # "None" and nothing is removed from lims
        tabs = lims.pop("tabs", None)

        # check that only one variable passed in
        if len(lims) != 1:
            raise pybamm.GeometryError("lims should only contain a single variable")

        ((spatial_var, spatial_lims),) = lims.items()
        return spatial_var, spatial_lims, tabs


class Uniform1DSubMesh(SubMesh1D):
    """
    A class to generate a uniform submesh on a 1D domain

    Parameters
    ----------
    lims : dict
        A dictionary that contains the limits of the spatial variables
    npts : dict
        A dictionary that contains the number of points to be used on each
        spatial variable. Note: the number of nodes (located at the cell centres)
        is npts, and the number of edges is npts+1.

    **Extends:"": :class:`pybamm.SubMesh1D`
    """

    def __init__(self, lims, npts):

        spatial_var, spatial_lims, tabs = self.read_lims(lims)
        npts = npts[spatial_var.id]

        edges = np.linspace(spatial_lims["min"], spatial_lims["max"], npts + 1)

        coord_sys = spatial_var.coord_sys

        super().__init__(edges, coord_sys=coord_sys, tabs=tabs)


class Exponential1DSubMesh(SubMesh1D):
    """
    A class to generate a submesh on a 1D domain in which the points are clustered
    close to one or both of boundaries using an exponential formula on the interval
    [a,b].

    If side is "left", the gridpoints are given by

    .. math::
        x_{k} = (b-a) + \\frac{\\exp{\\alpha k / N} - 1}{\\exp{\\alpha} - 1}} + a,

    for k = 1, ..., N, where N is the number of nodes.

    Is side is "right", the gridpoints are given by

    .. math::
        x_{k} = (b-a) + \\frac{\\exp{-\\alpha k / N} - 1}{\\exp{-\\alpha} - 1}} + a,

    for k = 1, ..., N.

    If side is "symmetric", the first half of the interval is meshed using the
    gridpoints

    .. math::
        x_{k} = (b/2-a) + \\frac{\\exp{\\alpha k / N} - 1}{\\exp{\\alpha} - 1}} + a,

    for k = 1, ..., N. The grid spacing is then reflected to contruct the grid
    on the full interval [a,b].

    In the above, alpha is a stretching factor. As the number of gridpoints tends
    to infinity, the ratio of the largest and smallest grid cells tends to exp(alpha).

    Parameters
    ----------
    lims : dict
        A dictionary that contains the limits of the spatial variables
    npts : dict
        A dictionary that contains the number of points to be used on each
        spatial variable. Note: the number of nodes (located at the cell centres)
        is npts, and the number of edges is npts+1.
    side : str, optional
        Whether the points are clustered near to the left or right boundary,
        or both boundaries. Can be "left", "right" or "symmetric". Default is
        "symmetric"
    stretch : float, optional
        The factor (alpha) which appears in the exponential. If side is "symmetric"
        then the default stretch is 1.15. If side is "left" or "right" then the
        default stretch is 2.3.

    **Extends:"": :class:`pybamm.SubMesh1D`
    """

    def __init__(self, lims, npts, side="symmetric", stretch=None):

        spatial_var, spatial_lims, tabs = self.read_lims(lims)
        a = spatial_lims["min"]
        b = spatial_lims["max"]
        npts = npts[spatial_var.id]
        coord_sys = spatial_var.coord_sys

        # Set stretch if not provided
        if not stretch:
            if side == "symmetric":
                stretch = 1.15
            elif side in ["left", "right"]:
                stretch = 2.3

        # Create edges accoriding to "side"
        if side == "left":
            ii = np.array(range(0, npts + 1))
            edges = (b - a) * (np.exp(stretch * ii / npts) - 1) / (
                np.exp(stretch) - 1
            ) + a

        elif side == "right":
            ii = np.array(range(0, npts + 1))
            edges = (b - a) * (np.exp(-stretch * ii / npts) - 1) / (
                np.exp(-stretch) - 1
            ) + a

        elif side == "symmetric":
            # Mesh half-interval [a, b/2]
            if npts % 2 == 0:
                ii = np.array(range(0, int((npts) / 2)))
            else:
                ii = np.array(range(0, int((npts + 1) / 2)))
            x_exp_left = (b / 2 - a) * (np.exp(stretch * ii / npts) - 1) / (
                np.exp(stretch) - 1
            ) + a

            # Refelct mesh
            x_exp_right = b * np.ones_like(x_exp_left) - (x_exp_left[::-1] - a)

            # Combine left and right halves of the mesh, adding a node at the
            # centre if npts is even (odd number of edges)
            if npts % 2 == 0:
                edges = np.concatenate((x_exp_left, [(a + b) / 2], x_exp_right))
            else:
                edges = np.concatenate((x_exp_left, x_exp_right))

        super().__init__(edges, coord_sys=coord_sys, tabs=tabs)


class Chebyshev1DSubMesh(SubMesh1D):
    """
    A class to generate a submesh on a 1D domain using Chebyshev nodes on the
    interval (a, b), given by

    .. math::
        x_{k} = \\frac{1}{2}(a+b) + \\frac{1}{2}(b-a) \\cos(\\frac{2k-1}{2N}\\pi),

    for k = 1, ..., N, where N is the number of nodes. Note: this mesh then
    appends the boundary edges, so that the mesh edges are given by

    .. math::
        a < x_{1} < ... < x_{N} < b.

    Parameters
    ----------
    lims : dict
        A dictionary that contains the limits of the spatial variables
    npts : dict
        A dictionary that contains the number of points to be used on each
        spatial variable. Note: the number of nodes (located at the cell centres)
        is npts, and the number of edges is npts+1.
    tabs : dict, optional
        A dictionary that contains information about the size and location of
        the tabs

    **Extends:"": :class:`pybamm.SubMesh1D`
    """

    def __init__(self, lims, npts, tabs=None):

        spatial_var, spatial_lims, tabs = self.read_lims(lims)
        npts = npts[spatial_var.id]

        # Create N Chebyshev nodes in the interval (a,b)
        N = npts - 1
        ii = np.array(range(1, N + 1))
        a = spatial_lims["min"]
        b = spatial_lims["max"]
        x_cheb = (a + b) / 2 + (b - a) / 2 * np.cos((2 * ii - 1) * np.pi / 2 / N)

        # Append the boundary nodes. Note: we need to flip the order the Chebyshev
        # nodes as they are created in descending order.
        edges = np.concatenate(([a], np.flip(x_cheb), [b]))
        coord_sys = spatial_var.coord_sys

        super().__init__(edges, coord_sys=coord_sys, tabs=tabs)


class UserSupplied1DSubMesh(SubMesh1D):
    """
    A class to generate a submesh on a 1D domain from a user supplied array of
    edges.

    Parameters
    ----------
    lims : dict
        A dictionary that contains the limits of the spatial variables
    npts : dict
        A dictionary that contains the number of points to be used on each
        spatial variable. Note: the number of nodes (located at the cell centres)
        is npts, and the number of edges is npts+1.
    edges : array_like
        The array of points which correspond to the edges of the mesh.

    **Extends:"": :class:`pybamm.SubMesh1D`
    """

    def __init__(self, lims, npts, edges=None):

        # raise error if no edges passed
        if edges is None:
            raise pybamm.GeometryError("User mesh requires parameter 'edges'")

        spatial_var, spatial_lims, tabs = self.read_lims(lims)
        npts = npts[spatial_var.id]

        # check that npts + 1 equals number of user-supplied edges
        if (npts + 1) != len(edges):
            raise pybamm.GeometryError(
                """User-suppled edges has should have length (npts + 1) but has length {}.
                 Number of points (npts) for domain {} is {}.""".format(
                    len(edges), spatial_var.domain, npts
                )
            )

        # check end points of edges agree with spatial_lims
        if edges[0] != spatial_lims["min"]:
            raise pybamm.GeometryError(
                """First entry of edges is {}, but should be equal to {}
                 for domain {}.""".format(
                    edges[0], spatial_lims["min"], spatial_var.domain
                )
            )
        if edges[-1] != spatial_lims["max"]:
            raise pybamm.GeometryError(
                """Last entry of edges is {}, but should be equal to {}
                for domain {}.""".format(
                    edges[-1], spatial_lims["max"], spatial_var.domain
                )
            )

        coord_sys = spatial_var.coord_sys

        super().__init__(edges, coord_sys=coord_sys, tabs=tabs)


class SpectralVolume1DSubMesh(SubMesh1D):
    """
    A class to subdivide any mesh to incorporate Chebyshev collocation
    Control Volumes. Note that the Spectral Volume method is optimized
    to only work with this submesh. The underlying theory could use any
    mesh with the right number of nodes, but in 1D the only sensible
    choice are the Chebyshev collocation points.

    Parameters
    ----------
    lims : dict
        A dictionary that contains the limits of the spatial variables
    npts : dict
        A dictionary that contains the number of points to be used on
        each spatial variable. Note: the number of nodes (located at the
        cell centres) is npts, and the number of edges is npts+1.
    order : int, optional
        The order of the Spectral Volume method that is to be used with
        this submesh. The default is 2, the same as the default for the
        SpectralVolume class. If the orders of the submesh and the
        Spectral Volume method don't match, the method will fail.
    **Extends:"": :class:`pybamm.SubMesh1D`
    """

    def __init__(self, lims, npts, edges=None, order=2):

        spatial_var, spatial_lims, tabs = self.read_lims(lims)
        npts = npts[spatial_var.id]

        # default: Spectral Volumes of equal size
        if edges is None:
            edges = np.linspace(spatial_lims["min"], spatial_lims["max"], npts + 1)
        # check that npts + 1 equals number of user-supplied edges
        elif (npts + 1) != len(edges):
            raise pybamm.GeometryError(
                "User-suppled edges should have length (npts + 1) but has len"
                "gth {}. Number of points (npts) for domain {} is {}.".format(
                    len(edges), spatial_var.domain, npts
                )
            )

        # check end points of edges agree with spatial_lims
        if edges[0] != spatial_lims["min"]:
            raise pybamm.GeometryError(
                """First entry of edges is {}, but should be equal to {}
                 for domain {}.""".format(
                    edges[0], spatial_lims["min"], spatial_var.domain
                )
            )
        if edges[-1] != spatial_lims["max"]:
            raise pybamm.GeometryError(
                """Last entry of edges is {}, but should be equal to {}
                for domain {}.""".format(
                    edges[-1], spatial_lims["max"], spatial_var.domain
                )
            )

        coord_sys = spatial_var.coord_sys

        cv_edges = np.array(
            [edges[0]]
            + [
                x
                for (a, b) in zip(edges[:-1], edges[1:])
                for x in np.flip(
                    a
                    + 0.5
                    * (b - a)
                    * (
                        1
                        + np.sin(
                            np.pi
                            * np.array(
                                [
                                    ((order + 1) - 1 - 2 * i) / (2 * (order + 1) - 2)
                                    for i in range(order + 1)
                                ]
                            )
                        )
                    )
                )[1:]
            ]
        )

        self.sv_edges = edges
        self.sv_nodes = (edges[:-1] + edges[1:]) / 2
        self.d_sv_edges = np.diff(self.sv_edges)
        self.d_sv_nodes = np.diff(self.sv_nodes)
        self.order = 2
        # The Control Volume edges and nodes are assigned to the
        # "edges" and "nodes" properties. This makes some of the
        # code of FiniteVolume directly applicable.
        super().__init__(cv_edges, coord_sys=coord_sys, tabs=tabs)



import os
import glob
import logging
import subprocess
from pathlib import Path
from platform import system, version
import wheel.bdist_wheel as orig
import site
import shutil

try:
    from setuptools import setup, find_packages, Extension
    from setuptools.command.install import install
except ImportError:
    from distutils.core import setup, find_packages
    from distutils.command.install import install

import CMakeBuild

default_lib_dir = (
    "" if system() == "Windows" else os.path.join(os.getenv("HOME"), ".local")
)

log_format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
logger = logging.getLogger("PyBaMM setup")

# To override the default severity of logging
logger.setLevel("INFO")

# Use FileHandler() to log to a file
logfile = os.path.join(os.path.dirname(os.path.abspath(__file__)), "setup.log")
file_handler = logging.FileHandler(logfile)
formatter = logging.Formatter(log_format)
file_handler.setFormatter(formatter)

# Add the file handler
logger.addHandler(file_handler)
logger.info("Starting PyBaMM setup")


class CustomInstall(install):
    """A custom install command to add 2 build options"""

    user_options = install.user_options + [
        ("suitesparse-root=", None, "suitesparse source location"),
        ("sundials-root=", None, "sundials source location"),
    ]

    def initialize_options(self):
        install.initialize_options(self)
        self.suitesparse_root = None
        self.sundials_root = None

    def finalize_options(self):
        install.finalize_options(self)
        if not self.suitesparse_root:
            self.suitesparse_root = default_lib_dir
        if not self.sundials_root:
            self.sundials_root = default_lib_dir

    def run(self):
        install.run(self)


class bdist_wheel(orig.bdist_wheel):
    """A custom install command to add 2 build options"""

    user_options = orig.bdist_wheel.user_options + [
        ("suitesparse-root=", None, "suitesparse source location"),
        ("sundials-root=", None, "sundials source location"),
    ]

    def initialize_options(self):
        orig.bdist_wheel.initialize_options(self)
        self.suitesparse_root = None
        self.sundials_root = None

    def finalize_options(self):
        orig.bdist_wheel.finalize_options(self)
        if not self.suitesparse_root:
            self.suitesparse_root = default_lib_dir
        if not self.sundials_root:
            self.sundials_root = default_lib_dir

    def run(self):
        orig.bdist_wheel.run(self)


def load_version():
    # Read version number from file
    try:
        root = os.path.abspath(os.path.dirname(__file__))
        with open(os.path.join(root, "pybamm", "version"), "r") as f:
            version = f.read().strip().split(",")
        return ".".join(["{:02d}".format(int(x)) for x in version])
    except Exception as e:
        raise RuntimeError("Unable to read version number (" + str(e) + ").")


def compile_KLU():
    # Return whether or not the KLU extension should be compiled.
    # Return True if:
    # - Not running on Windows AND
    # - CMake is found AND
    # - The pybind11 directory is found in the PyBaMM project directory
    CMakeFound = True
    PyBind11Found = True
    windows = (not system()) or system() == "Windows"

    msg = "Running on Windows" if windows else "Not running on windows"
    logger.info(msg)

    try:
        subprocess.run(["cmake", "--version"])
        logger.info("Found CMake.")
    except OSError:
        CMakeFound = False
        logger.info("Could not find CMake. Skipping compilation of KLU module.")

    pybamm_project_dir = os.path.dirname(os.path.abspath(__file__))
    pybind11_dir = os.path.join(pybamm_project_dir, "pybind11")
    try:
        open(os.path.join(pybind11_dir, "tools", "pybind11Tools.cmake"))
        logger.info("Found pybind11 directory ({})".format(pybind11_dir))
    except FileNotFoundError:
        PyBind11Found = False
        msg = (
            "Could not find PyBind11 directory ({})."
            " Skipping compilation of KLU module.".format(pybind11_dir)
        )
        logger.info(msg)

    return CMakeFound and PyBind11Found


# Build the list of package data files to be included in the PyBaMM package.
# These are mainly the parameter files located in the input/parameters/ subdirectories.
pybamm_data = []
for file_ext in ["*.csv", "*.py", "*.md", "*.txt"]:
    # Get all the files ending in file_ext in pybamm/input dir.
    # list_of_files = [
    #    'pybamm/input/drive_cycles/car_current.csv',
    #    'pybamm/input/drive_cycles/US06.csv',
    # ...
    list_of_files = glob.glob("pybamm/input/**/" + file_ext, recursive=True)

    # Add these files to pybamm_data.
    # The path must be relative to the package dir (pybamm/), so
    # must process the content of list_of_files to take out the top
    # pybamm/ dir, i.e.:
    # ['input/drive_cycles/car_current.csv',
    #  'input/drive_cycles/US06.csv',
    # ...
    pybamm_data.extend(
        [os.path.join(*Path(filename).parts[1:]) for filename in list_of_files]
    )
pybamm_data.append("./version")
pybamm_data.append("./CITATIONS.txt")
pybamm_data.append("./plotting/pybamm.mplstyle")
pybamm_data.append("../CMakeBuild.py")

idaklu_ext = Extension("pybamm.solvers.idaklu", ["pybamm/solvers/c_solvers/idaklu.cpp"])
ext_modules = [idaklu_ext] if compile_KLU() else []

jax_dependencies = []
if not (system() == "Windows" or (system() == "Darwin" and "ARM64" in version())):
    jax_dependencies = ["jax==0.2.12", "jaxlib==0.1.70"]


# Load text for description and license
with open("README.md", encoding="utf-8") as f:
    readme = f.read()

setup(
    name="pybamm",
    version=load_version(),
    description="Python Battery Mathematical Modelling.",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/pybamm-team/PyBaMM",
    packages=find_packages(include=("pybamm", "pybamm.*")),
    ext_modules=ext_modules,
    cmdclass={
        "build_ext": CMakeBuild.CMakeBuild,
        "bdist_wheel": bdist_wheel,
        "install": CustomInstall,
    },
    package_data={"pybamm": pybamm_data},
    # Python version
    python_requires=">=3.7,<3.10",
    # List of dependencies
    install_requires=[
        "numpy>=1.16",
        "scipy>=1.3",
        "pandas>=0.24",
        "anytree>=2.4.3",
        "autograd>=1.2",
        "scikit-fem>=0.2.0",
        "casadi>=3.5.0",
        "imageio>=2.9.0",
        *jax_dependencies,
        "jupyter",  # For example notebooks
        "pybtex",
        "sympy==1.8",
        # Note: Matplotlib is loaded for debug plots, but to ensure pybamm runs
        # on systems without an attached display, it should never be imported
        # outside of plot() methods.
        # Should not be imported
        "matplotlib>=2.0",
    ],
    extras_require={
        "docs": ["sphinx>=1.5", "guzzle-sphinx-theme"],  # For doc generation
        "dev": [
            "flake8>=3",  # For code style checking
            "black",  # For code style auto-formatting
        ],
    },
    entry_points={
        "console_scripts": [
            "pybamm_edit_parameter = pybamm.parameters_cli:edit_parameter",
            "pybamm_add_parameter = pybamm.parameters_cli:add_parameter",
            "pybamm_rm_parameter = pybamm.parameters_cli:remove_parameter",
            "pybamm_install_odes = pybamm.install_odes:main",
        ]
    },
)

# pybtex adds a folder "tests" to the site packages, so we manually remove this
path_to_sitepackages = site.getsitepackages()[0]
path_to_tests_dir = os.path.join(path_to_sitepackages, "tests")
if os.path.exists(path_to_tests_dir):
    shutil.rmtree(path_to_tests_dir)
